import unittest

import torch
from hamcrest import assert_that, equal_to

import kerosene.nn.functional as F


class TestJensenShannonDivergence(unittest.TestCase):
    PROB_DIST1, PROB_DIST2, PROB_DIST3 = [1 / 2, 1 / 2, 0], [0, 1 / 10, 9 / 10], [1 / 3, 1 / 3, 1 / 3]

    def test_should_compute_jensen_shannon_divergence(self):
        prob_distributions = torch.tensor([[self.PROB_DIST1, self.PROB_DIST2, self.PROB_DIST3]])
        expected_results = torch.tensor([0.378889])

        assert_that(F.js_div(prob_distributions), equal_to(expected_results))

    def test_should_compute_jensen_shannon_divergence_of_same_distribution(self):
        prob_distributions = torch.tensor([[self.PROB_DIST1, self.PROB_DIST1, self.PROB_DIST1]])
        expected_results = torch.tensor([0.0])

        assert_that(F.js_div(prob_distributions), equal_to(expected_results))



import torch.nn.functional as F
from torch.nn import Module, Linear, Dropout2d, Conv2d


class SimpleNet(Module):
    def __init__(self):
        super().__init__()
        self.conv1 = Conv2d(1, 10, kernel_size=5)
        self.conv2 = Conv2d(10, 20, kernel_size=5)
        self.conv2_drop = Dropout2d()
        self.fc1 = Linear(320, 50)
        self.fc2 = Linear(50, 10)

    def forward(self, x):
        x = F.relu(F.max_pool2d(self.conv1(x), 2))
        x = F.relu(F.max_pool2d(self.conv2_drop(self.conv2(x)), 2))
        x = x.view(-1, 320)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, training=self.training)
        x = self.fc2(x)
        return x


from datetime import datetime
from enum import Enum


class Phase(Enum):
    TRAINING = "Training"
    VALIDATION = "Validation"
    TEST = "Test"
    ALL = [TRAINING, VALIDATION, TEST]

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, Phase):
            return self.value == other.value

    def __hash__(self):
        return hash(self.value)


class Frequency(Enum):
    STEP = "Step"
    EPOCH = "Epoch"
    PHASE = "Phase"

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, Frequency):
            return self.value == other.value

    def __hash__(self):
        return hash(self.value)


class Moment(object):
    def __init__(self, iteration, frequency: Frequency, phase: Phase, time: datetime = None):
        self._datetime = datetime.now() if time is None else time
        self._frequency = frequency
        self._iteration = iteration
        self._phase = phase

    @property
    def datetime(self):
        return self._datetime

    @datetime.setter
    def datetime(self, value):
        self._datetime = value

    @property
    def frequency(self):
        return self._frequency

    @frequency.setter
    def frequency(self, value):
        self._frequency = value

    @property
    def iteration(self):
        return self._iteration

    @iteration.setter
    def iteration(self, value):
        self._iteration = value

    @property
    def phase(self):
        return self._phase

    @phase.setter
    def phase(self, value):
        self._phase = value


class BaseEvent(Enum):
    def __init__(self, value):
        self._value_ = value

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, BaseEvent):
            return self.value == other.value

    def __hash__(self):
        return hash(self.value)


class TemporalEvent(object):
    def __init__(self, event: BaseEvent, moment: Moment):
        self._event = event
        self._moment = moment

    @property
    def event(self):
        return self._event

    @property
    def frequency(self):
        return self._moment.frequency

    @property
    def phase(self):
        return self._moment.phase

    @property
    def datetime(self):
        return self._moment.datetime

    @property
    def iteration(self):
        return self._moment.iteration

    def __eq__(self, other):
        if isinstance(other, BaseEvent):
            return self.event == other
        elif isinstance(other, TemporalEvent):
            return self._event == other.event and self.frequency == other.frequency and \
                   self.phase == other.phase and self.datetime == other.datetime and self.iteration == other.iteration

    def __str__(self):
        return str(self.event)


class BaseVariable(Enum):
    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, BaseVariable):
            return self.value == other.value

    def __hash__(self):
        return hash(self.value)


class Monitor(BaseVariable):
    LOSS = "loss"
    METRICS = "metrics"

    TRAIN_LOSS = "train_loss"
    VALID_LOSS = "valid_loss"
    TEST_LOSS = "test_loss"
    TRAIN_METRICS = "train_metrics"
    VALID_METRICS = "valid_metrics"
    TEST_METRICS = "test_metrics"

    def is_loss(self):
        return "loss" in self.value

    def is_metrics(self):
        return "metrics" in self.value


class MonitorMode(Enum):
    MIN = -1
    MAX = 1

    def __str__(self):
        return self.value



from kerosene.events import TemporalEvent, Moment, Frequency, Phase, BaseEvent
from kerosene.training import Status


class Event(BaseEvent):
    ON_TRAINING_BEGIN = "training_begin"
    ON_TRAINING_END = "training_end"
    ON_VALID_BEGIN = "valid_begin"
    ON_VALID_END = "valid_end"
    ON_TEST_BEGIN = "test_begin"
    ON_TEST_END = "test_end"
    ON_EPOCH_BEGIN = "epoch_begin"
    ON_EPOCH_END = "epoch_end"
    ON_TRAIN_EPOCH_BEGIN = "train_epoch_begin"
    ON_TRAIN_EPOCH_END = "train_epoch_end"
    ON_VALID_EPOCH_BEGIN = "valid_epoch_begin"
    ON_VALID_EPOCH_END = "valid_epoch_end"
    ON_TEST_EPOCH_BEGIN = "test_epoch_begin"
    ON_TEST_EPOCH_END = "test_epoch_end"
    ON_BATCH_BEGIN = "batch_begin"
    ON_TRAIN_BATCH_BEGIN = "train_batch_begin"
    ON_TRAIN_BATCH_END = "train_batch_end"
    ON_VALID_BATCH_BEGIN = "valid_batch_begin"
    ON_VALID_BATCH_END = "valid_batch_end"
    ON_TEST_BATCH_BEGIN = "test_batch_begin"
    ON_TEST_BATCH_END = "test_batch_end"
    ON_BATCH_END = "batch_end"
    ON_FINALIZE = "finalizing"

    def __call__(self, moment: Moment):
        return TemporalEvent(self, moment)


class BatchEventPublisherMixin(object):
    @property
    def iteration_and_phase(self):
        if self.status is Status.TRAINING:
            iteration = self.current_train_step
            phase = Phase.TRAINING
        elif self.status is Status.VALIDATING:
            iteration = self.current_valid_step
            phase = Phase.VALIDATION
        else:
            iteration = self.current_test_step
            phase = Phase.TEST

        return iteration, phase

    def on_batch_begin(self):
        pass

    def on_batch_end(self):
        pass

    def on_train_batch_begin(self):
        pass

    def on_train_batch_end(self):
        pass

    def on_valid_batch_begin(self):
        pass

    def on_valid_batch_end(self):
        pass

    def on_test_batch_begin(self):
        pass

    def on_test_batch_end(self):
        pass

    def _on_batch_begin(self):
        self.on_batch_begin()

        iteration, phase = self.iteration_and_phase

        self.fire(Event.ON_BATCH_BEGIN(Moment(iteration, Frequency.STEP, phase)))

    def _on_batch_end(self):
        self.on_batch_end()

        iteration, phase = self.iteration_and_phase

        self.fire(Event.ON_BATCH_END(Moment(iteration, Frequency.STEP, phase)), self.step_monitors(phase))

    def _on_train_batch_begin(self):
        self.on_train_batch_begin()
        self.fire(Event.ON_TRAIN_BATCH_BEGIN(Moment(self.current_train_step, Frequency.STEP, Phase.TRAINING)))

    def _on_train_batch_end(self):
        self.on_train_batch_end()
        self.fire(Event.ON_TRAIN_BATCH_END(Moment(self.current_train_step, Frequency.STEP, Phase.TRAINING)),
                  self.step_monitors(Phase.TRAINING))

    def _on_valid_batch_begin(self):
        self.on_valid_batch_begin()
        self.fire(Event.ON_VALID_BATCH_BEGIN(Moment(self.current_valid_step, Frequency.STEP, Phase.VALIDATION)))

    def _on_valid_batch_end(self):
        self.on_valid_batch_begin()
        self.fire(Event.ON_VALID_BATCH_END(Moment(self.current_valid_step, Frequency.STEP, Phase.VALIDATION)),
                  self.step_monitors(Phase.VALIDATION))

    def _on_test_batch_begin(self):
        self.on_test_batch_begin()
        self.fire(Event.ON_TEST_BATCH_BEGIN(Moment(self.current_test_step, Frequency.STEP, Phase.TEST)))

    def _on_test_batch_end(self):
        self.on_test_batch_end()
        self.fire(Event.ON_TEST_BATCH_END(Moment(self.current_test_step, Frequency.STEP, Phase.TEST)),
                  self.step_monitors(Phase.TEST))


class EpochEventPublisherMixin(object):
    @property
    def epoch_and_phase(self):
        if self.status is Status.TRAINING:
            phase = Phase.TRAINING
        elif self.status is Status.VALIDATING:
            phase = Phase.VALIDATION
        else:
            phase = Phase.TEST

        return self.epoch, phase

    def on_epoch_begin(self):
        pass

    def on_epoch_end(self):
        pass

    def on_train_epoch_begin(self):
        pass

    def on_train_epoch_end(self):
        pass

    def on_valid_epoch_begin(self):
        pass

    def on_valid_epoch_end(self):
        pass

    def on_test_epoch_begin(self):
        pass

    def on_test_epoch_end(self):
        pass

    def _on_epoch_begin(self):
        self._reset_model_trainers()
        self.on_epoch_begin()

        epoch, phase = self.epoch_and_phase

        self.fire(Event.ON_EPOCH_BEGIN(Moment(epoch, Frequency.EPOCH, phase)))

    def _on_epoch_end(self):
        self.on_epoch_end()

        epoch, phase = self.epoch_and_phase

        self.fire(Event.ON_EPOCH_END(Moment(epoch, Frequency.EPOCH, phase)), self.epoch_monitors(phase))

    def _on_train_epoch_begin(self):
        self._status = Status.TRAINING
        self.on_train_epoch_begin()
        self.fire(Event.ON_TRAIN_EPOCH_BEGIN(Moment(self.epoch, Frequency.EPOCH, Phase.TRAINING)))

    def _on_train_epoch_end(self):
        self.on_train_epoch_end()
        self.fire(Event.ON_TRAIN_EPOCH_END(Moment(self.epoch, Frequency.EPOCH, Phase.TRAINING)),
                  self.epoch_monitors(Phase.TRAINING))

    def _on_valid_epoch_begin(self):
        self.on_valid_epoch_begin()
        self.fire(Event.ON_VALID_EPOCH_BEGIN(Moment(self.epoch, Frequency.EPOCH, Phase.VALIDATION)))

    def _on_valid_epoch_end(self):
        self._current_valid_batch = 0
        self.on_valid_epoch_end()
        self.fire(Event.ON_VALID_EPOCH_END(Moment(self.epoch, Frequency.EPOCH, Phase.VALIDATION)),
                  self.epoch_monitors(Phase.VALIDATION))

    def _on_test_epoch_begin(self):
        self.on_test_epoch_begin()
        self.fire(Event.ON_TEST_EPOCH_BEGIN(Moment(self.epoch, Frequency.EPOCH, Phase.TEST)))

    def _on_test_epoch_end(self):
        self._current_test_batch = 0
        self.on_test_epoch_end()
        self.fire(Event.ON_TEST_EPOCH_END(Moment(self.epoch, Frequency.STEP, Phase.TEST)),
                  self.epoch_monitors(Phase.TEST))


class TrainingPhaseEventPublisherMixin(object):
    def on_training_begin(self):
        pass

    def on_training_end(self):
        pass

    def on_valid_begin(self):
        pass

    def on_valid_end(self):
        pass

    def on_test_begin(self):
        pass

    def on_test_end(self):
        pass

    def finalize(self):
        pass

    def _on_training_begin(self):
        self.on_training_begin()
        self._status = Status.TRAINING
        self.fire(Event.ON_TRAINING_BEGIN(Moment(0, Frequency.PHASE, Phase.TRAINING)))

    def _on_training_end(self):
        self.on_training_end()
        self.scheduler_step()
        self.fire(Event.ON_TRAINING_END(Moment(0, Frequency.PHASE, Phase.TRAINING)))

    def _on_valid_begin(self):
        self.on_valid_begin()
        self._status = Status.VALIDATING
        self.fire(Event.ON_VALID_BEGIN(Moment(0, Frequency.PHASE, Phase.VALIDATION)))

    def _on_valid_end(self):
        self.on_valid_end()
        self.fire(Event.ON_VALID_END(Moment(0, Frequency.PHASE, Phase.VALIDATION)))

    def _on_test_begin(self):
        self.on_test_begin()
        self._status = Status.TESTING
        self.fire(Event.ON_TEST_BEGIN(Moment(0, Frequency.PHASE, Phase.TEST)))

    def _on_test_end(self):
        self.on_test_end()
        self.fire(Event.ON_TEST_END(Moment(0, Frequency.PHASE, Phase.TEST)))

    def _finalize(self):
        for model_trainer in self.model_trainers:
            model_trainer.finalize()
        self.finalize()
        self.status = Status.FINALIZED



class UnsupportedEventException(Exception):

    def __init__(self, supported_events, unsupported_event):
        super(UnsupportedEventException, self).__init__(
            "Unsupported event provided ({}). Only {} are permitted".format(str(unsupported_event),
                                                                            [str(event) for event in
                                                                             supported_events]))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import division

from abc import ABC
from typing import Optional

import torch
from torch import Tensor, nn
from torch.nn.parallel.distributed import DistributedDataParallel as DDP
from torch.optim import Optimizer

from kerosene.utils.devices import on_multiple_gpus, get_devices

try:
    from apex import amp
    from apex.parallel import DistributedDataParallel as ApexDDP

    APEX_AVAILABLE = True
except ModuleNotFoundError:
    APEX_AVAILABLE = False


class ApexLoss(object):
    def __init__(self, loss_id: int, loss: Tensor, optimizer: Optimizer):
        super().__init__()
        self._loss_id = loss_id
        self._loss = loss
        self._optimizer = optimizer

    @property
    def loss_id(self):
        return self._loss_id

    @property
    def loss(self):
        return self._loss

    def backward(self, gradient: Optional[Tensor] = None, retain_graph=False,
                 create_graph=False) -> None:
        if APEX_AVAILABLE:
            with amp.scale_loss(self._loss, self._optimizer, loss_id=self._loss_id) as scaled_loss:
                scaled_loss.backward(gradient, retain_graph, create_graph)
        else:
            self._loss.backward(gradient, retain_graph, create_graph)

    def cpu(self):
        return ApexLoss(self._loss_id, self._loss.cpu(), self._optimizer)

    def detach(self):
        return ApexLoss(self._loss_id, self._loss.detach(), self._optimizer)

    def float(self):
        return ApexLoss(self._loss_id, self._loss.float(), self._optimizer)

    def numpy(self):
        return self._loss.numpy()

    def mean(self):
        return ApexLoss(self._loss_id, torch.mean(self._loss), self._optimizer)
        
    def item(self):
    	return self._loss.item()

    def __add__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.add(self._loss, other), self._optimizer)
        elif isinstance(other, ApexLoss):
            loss = ApexLoss(self._loss_id, torch.add(self._loss, other.loss), self._optimizer)
        else:
            raise NotImplementedError("Cannot add an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __mul__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.mul(self._loss, other), self._optimizer)
        elif isinstance(other, ApexLoss):
            loss = ApexLoss(self._loss_id, torch.mul(self._loss, other.loss), self._optimizer)
        else:
            raise NotImplementedError("Cannot mul an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __rmul__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.mul(self._loss, other), self._optimizer)
        else:
            raise NotImplementedError("Cannot rmul an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __truediv__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.div(self._loss, other), self._optimizer)
        elif isinstance(other, ApexLoss):
            loss = ApexLoss(self._loss_id, torch.div(self._loss, other.loss), self._optimizer)
        else:
            raise NotImplementedError("Cannot truediv an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __rtruediv__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.div(other, self._loss), self._optimizer)
        else:
            raise NotImplementedError("Cannot rtruediv an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __sub__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.sub(self._loss, other), self._optimizer)
        elif isinstance(other, ApexLoss):
            loss = ApexLoss(self._loss_id, torch.sub(self._loss, other.loss), self._optimizer)
        else:
            raise NotImplementedError(
                "Cannot substract an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __rsub__(self, other):
        if isinstance(other, (Tensor, int, float)):
            loss = ApexLoss(self._loss_id, torch.sub(other, self._loss), self._optimizer)
        else:
            raise NotImplementedError(
                "Cannot substract an element of type: {} to an ApexLoss.".format(str(type(other))))
        return loss

    def __eq__(self, other):
        if isinstance(other, Tensor):
            is_equal = torch.all(torch.eq(self._loss, other))
        elif isinstance(other, ApexLoss):
            is_equal = torch.all(torch.eq(self._loss, other.loss))
        else:
            raise NotImplementedError("Cannot compare an element of type: {} to an ApexLoss.".format(str(type(other))))
        return is_equal


class ApexModule(ABC, nn.Module):
    def __init__(self, model, optimizer, amp_id=0, use_amp=True):
        super().__init__()
        self._amp_id = amp_id
        self._use_amp = use_amp
        self._model = model
        self._optimizer = optimizer

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def amp_id(self):
        return self._amp_id

    @property
    def use_amp(self):
        return self._use_amp

    def initialize(self, amp_id: int, num_losses: int, use_amp: bool, amp_opt_level: str, device: torch.device):
        self._amp_id = amp_id
        self._use_amp = use_amp

        if APEX_AVAILABLE and self._use_amp:
            self._model, self._optimizer = amp.initialize(
                self._model, self._optimizer, opt_level=amp_opt_level, num_losses=num_losses)
            if on_multiple_gpus(get_devices()):
                self._model = ApexDDP(self._model, delay_allreduce=True)
        if not APEX_AVAILABLE and on_multiple_gpus(get_devices()):
            self._model = DDP(self._model, device_ids=[device])



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Union

from torch import optim
from torch.optim import Optimizer

try:
    from apex.optimizers import FusedSGD, FusedAdam

    APEX_AVAILABLE = True
except ModuleNotFoundError:
    APEX_AVAILABLE = False
except ImportError as e:
    APEX_AVAILABLE = False
    print("Unable to import apex optimizers please upgrade your apex version".format(e))


class OptimizerType(Enum):
    FusedSGD = "FusedSGD"
    FusedAdam = "FusedAdam"
    SGD = "SGD"
    Adam = "Adam"
    Adagrad = "Adagrad"
    Adadelta = "Adadelta"
    SparseAdam = "SparseAdam"
    Adamax = "Adamax"
    Rprop = "Rprop"
    RMSprop = "RMSprop"
    ASGD = "ASGD"

    def __str__(self):
        return self.value


class OptimizerFactory(object):

    def __init__(self):
        self._optimizers = {
            "Adam": optim.Adam,
            "SGD": optim.SGD,
        }

        if APEX_AVAILABLE:
            self.register("FusedSGD", FusedSGD)
            self.register("FusedAdam", FusedAdam)

    def create(self, optimizer_type: Union[str, OptimizerType], model_params, params):
        return self._optimizers[str(optimizer_type)](model_params, **params) if params is not None else \
            self._optimizers[str(optimizer_type)](model_params)

    def register(self, function: str, creator: Optimizer):
        """
        Add a new activation layer.
        Args:
           function (str): Activation layer name.
           creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom optimizer function.
        """
        self._optimizers[function] = creator



import unittest

from hamcrest import *

from kerosene.nn.utils.gradients import GradientClippingStrategyFactory, GradientClippingStrategyType, \
    GradientClippingStrategy


class TestGradientClippingStrategyFactory(unittest.TestCase):
    def setUp(self) -> None:
        pass

    def test_should_return_a_gradient_norm_clipping_strategy(self):
        strategy = GradientClippingStrategyFactory().create(GradientClippingStrategyType.Norm,
                                                            {"max_norm": 1.0, "norm_type": 2})

        assert_that(strategy, instance_of(GradientClippingStrategy))

    def test_should_return_a_gradient_value_clipping_strategy(self):
        strategy = GradientClippingStrategyFactory().create(GradientClippingStrategyType.Value, {"clip_value": 1.0})

        assert_that(strategy, instance_of(GradientClippingStrategy))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import abc
from enum import Enum
from typing import Union, Callable, Iterable

import torch
from torch.nn.utils import clip_grad_norm_, clip_grad_value_


class GradientClippingStrategyType(Enum):
    Value = "value"
    Norm = "norm"

    def __str__(self):
        return self.value


class GradientClippingStrategy(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def clip(self, model_parameters: Union[Iterable[torch.Tensor], torch.Tensor]):
        raise NotImplementedError()


class GradientNormClipping(GradientClippingStrategy):

    def __init__(self, max_norm: float, norm_type: Union[int, float] = 2):
        """"
        Gradient norm clipping strategy.

        Args:
            max_norm (float or int): max norm of the gradients
            norm_type (float or int): type of the used p-norm. Can be ``'inf'`` for infinity norm.
        """
        self._max_norm = max_norm
        self._norm_type = norm_type

    def clip(self, model_parameters):
        clip_grad_norm_(model_parameters, self._max_norm, self._norm_type)


class GradientValueClipping(GradientClippingStrategy):

    def __init__(self, clip_value: Union[int, float]):
        """"
        Gradient value clipping strategy.

        Args:
           clip_value (float or int): maximum allowed value of the gradients. The gradients are clipped in the range
            :math:`\left[\text{-clip\_value}, \text{clip\_value}\right]`
        """
        self._clip_value = clip_value

    def clip(self, model_parameters: Union[Iterable[torch.Tensor], torch.Tensor]):
        clip_grad_value_(model_parameters, self._clip_value)


class GradientClippingStrategyFactory(object):
    def __init__(self):
        self._clipping_strategies = {
            "value": GradientValueClipping,
            "norm": GradientNormClipping
        }

    def create(self, clipping_strategy: Union[str, GradientClippingStrategyType], params):
        if clipping_strategy is not None:
            return self._clipping_strategies[str(clipping_strategy)](**params)
        else:
            return None

    def register(self, function: str, creator: Callable):
        self._clipping_strategies[function] = creator



import logging
import os
from typing import Callable

import torch

from kerosene.events import MonitorMode, TemporalEvent
from kerosene.events.handlers.base_monitor_watcher import MonitorWatcher, MonitorPatienceExceeded
from kerosene.training.events import Event
from kerosene.training.trainers import Trainer
from kerosene.utils.constants import CHECKPOINT_EXT
from kerosene.utils.files import should_create_dir


class Checkpoint(MonitorWatcher):
    LOGGER = logging.getLogger("Checkpoint")
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END]

    def __init__(self, path, monitor_fn: Callable, delta: float, mode: MonitorMode):
        super(Checkpoint, self).__init__(monitor_fn, mode, delta, patience=0, supported_events=self.SUPPORTED_EVENTS)
        self._path = path

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        if self.should_handle(event):
            for model_trainer in trainer.model_trainers:
                try:
                    value = self._monitor_fn(model_trainer)
                    self.watch(model_trainer.name, value)
                except MonitorPatienceExceeded as e:
                    self._save_model(model_trainer.name, model_trainer.model_state, event.iteration)
                    self._save_optimizer(model_trainer.name, model_trainer.optimizer_state)

    def _save_model(self, model_name, model_state, epoch_num):
        if should_create_dir(self._path, model_name):
            os.makedirs(os.path.join(self._path, model_name))
        torch.save({"model_state_dict": model_state, "epoch_num": epoch_num},
                   os.path.join(self._path, model_name, model_name + CHECKPOINT_EXT))

    def _save_optimizer(self, model_name, optimizer_state):
        if should_create_dir(self._path, model_name):
            os.makedirs(os.path.join(self._path, model_name))
        torch.save({"optimizer_state_dict": optimizer_state},
                   os.path.join(self._path, model_name, "{}_optimizer{}".format(model_name, CHECKPOINT_EXT)))



import logging
import unittest

from mockito import verify, spy

from kerosene.events import Phase
from kerosene.events.handlers.console import PrintTrainingStatus, StatusConsoleColorPalette


class ConsoleHandlerTest(unittest.TestCase):

    def setUp(self) -> None:
        logging.basicConfig(level=logging.INFO)
        self._logger_mock = spy(logging.getLogger())

    def test_print_status_with_colors(self):
        expected_string = "\nCurrent states: [92mTraining[0m |  Epoch: 0 | Training step: 0 | Validation step: 0 | Test step: 0\n"
        handler = PrintTrainingStatus(colors=StatusConsoleColorPalette.DEFAULT)
        handler.LOGGER = self._logger_mock
        handler.print_status(Phase.TRAINING, 0, 0, 0, 0)
        verify(self._logger_mock).info(expected_string)

    def test_print_status_without_colors(self):
        expected_string = "\nCurrent states: Training\x1b[0m |  Epoch: 0 | Training step: 0 | Validation step: 0 | Test step: 0\n"
        handler = PrintTrainingStatus()
        handler.LOGGER = self._logger_mock
        handler.print_status(Phase.TRAINING, 0, 0, 0, 0)
        verify(self._logger_mock).info(expected_string)



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import ABC, abstractmethod
from typing import List, Union

from kerosene.events import TemporalEvent, BaseEvent
from kerosene.events.exceptions import UnsupportedEventException


class EventHandler(ABC):

    def __init__(self, supported_events: List[Union[BaseEvent, TemporalEvent]] = None, every=1):
        self._every = every
        self._supported_events = supported_events

    @property
    def every(self):
        return self._every

    @property
    def supported_events(self):
        return self._supported_events

    def should_handle(self, event: TemporalEvent):
        if (self.supported_events is not None) and (event not in self.supported_events):
            raise UnsupportedEventException(self.supported_events, event)

        if iter == 0 and self._every != 1:
            return False
        else:
            return event.iteration % self._every == 0

    @abstractmethod
    def __call__(self, temporal_event: TemporalEvent, monitors, sender):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging

import torch
import yaml

from kerosene.configs.configs import ModelConfiguration, TrainerConfiguration


class CustomYamlParser(object):

    def __init__(self):
        yaml.SafeLoader.add_constructor(u"!torch/tensor", CustomYamlParser.parse_tensor)
        yaml.SafeLoader.add_constructor(u"!python/tuple", CustomYamlParser.parse_tuple)

    @staticmethod
    def safe_load(file):
        return yaml.safe_load(file)

    @staticmethod
    def parse_tensor(loader, node):
        value = loader.construct_sequence(node, deep=True)
        tensor = torch.Tensor().new_tensor(value)
        return tensor

    @staticmethod
    def parse_tuple(loader, node):
        value = loader.construct_sequence(node, deep=True)
        tuple_ = tuple(value)
        return tuple_


class YamlConfigurationParser(object):
    LOGGER = logging.getLogger("YamlConfigurationParser")

    @staticmethod
    def parse(config_file_path):
        with open(config_file_path, 'r') as config_file:
            try:
                config = CustomYamlParser().safe_load(config_file)

                model_trainer_configs = list(
                    map(lambda model_name: ModelConfiguration.from_dict(model_name,
                                                                        config["models"][model_name]),
                        config["models"]))

                model_trainer_configs = model_trainer_configs if len(model_trainer_configs) > 1 else \
                    model_trainer_configs[0]
                training_config = TrainerConfiguration(config['training'])
                return model_trainer_configs, training_config
            except yaml.YAMLError as e:
                YamlConfigurationParser.LOGGER.warning(
                    "Unable to read the training config file: {} with error {}".format(config_file_path, e))

    @staticmethod
    def parse_section(config_file_path, yml_tag):
        with open(config_file_path, 'r') as config_file:
            try:
                config = CustomYamlParser().safe_load(config_file)

                return config[yml_tag]
            except yaml.YAMLError as e:
                YamlConfigurationParser.LOGGER.warning(
                    "Unable to read the training config file: {} with error {}".format(config_file_path, e))



import torch
import torch.nn.functional as F


def js_div(inputs):
    js_divs = []
    for dist in range(inputs.size(1)):
        js_divs.append(
            F.kl_div(torch.mean(inputs, dim=1, keepdim=True).log(), inputs[:, dist].unsqueeze(0), reduction='sum'))

    return torch.tensor(js_divs).mean()



import unittest

import torch
from hamcrest import *

from kerosene.utils.devices import on_multiple_gpus, on_single_gpu, get_devices


class DeviceTests(unittest.TestCase):

    def setUp(self) -> None:
        self._multiple_gpus_devices = [torch.device("cuda:0"), torch.device("cuda:1")]
        self._single_gpu_device = [torch.device("cuda:0")]
        self._single_cpu_device = [torch.device("cpu")]

    def test_on_single_gpu_should_return_true_with_single_GPU_device(self):
        assert_that(on_single_gpu(self._single_gpu_device), is_(True))

    def test_on_single_gpu_should_return_false_with_multiple_GPU_devices(self):
        assert_that(on_single_gpu(self._multiple_gpus_devices), is_(False))

    def test_on_multiple_gpu_should_return_false_with_single_GPU_device(self):
        assert_that(on_multiple_gpus(self._single_gpu_device), is_(False))

    def test_on_multiple_gpu_should_return_true_with_multiple_GPU_devices(self):
        assert_that(on_multiple_gpus(self._multiple_gpus_devices), is_(True))

    def test_get_devices_should_return_at_least_one_cuda_enabled_device(self):
        if torch.cuda.is_available():
            assert_that(get_devices(), has_item(torch.device("cuda:0")))
            assert_that(get_devices(), has_length(torch.cuda.device_count()))
            assert_that(get_devices(), is_(
                [torch.device("cuda:{}".format(id)) for id in range(int(torch.cuda.device_count()))]))
        else:
            assert_that(get_devices(), is_([torch.device("cpu")]))



import unittest

from mockito import mock, spy

from kerosene.events.handlers.visdom import PlotLosses
from kerosene.loggers.visdom.visdom import VisdomLogger
from tests.base_test import BaseTestEnvironment


class VisdomTest(BaseTestEnvironment):

    def setUp(self):
        self._visdom_logger_mock = mock(VisdomLogger)
        self._plot_losses_handler = spy(PlotLosses)

    def test_should_plot_losses(self):
        pass



from abc import ABC
from typing import Dict, List, Union

from kerosene.events import MonitorMode, BaseEvent, TemporalEvent
from kerosene.events.handlers.base_handler import EventHandler


class MonitorPatienceExceeded(Exception):
    pass


class MonitorInspection(object):
    def __init__(self, value=0, inspection_num=0):
        self._value = value
        self._inspection_num = inspection_num

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value

    @property
    def inspection_num(self):
        return self._inspection_num

    def add_inspection(self):
        self._inspection_num = self._inspection_num + 1

    def reset_inspection_num(self):
        self._inspection_num = 0
        return self

    def with_value(self, value):
        self._value = value
        return self


class MonitorWatcher(EventHandler, ABC):
    def __init__(self, monitor_fn, mode: MonitorMode, min_delta, patience, every=1,
                 supported_events: List[Union[BaseEvent, TemporalEvent]] = None):
        super().__init__(supported_events, every)
        self._monitor_fn = monitor_fn
        self._mode = mode
        self._min_delta = min_delta
        self._patience = patience
        self._monitor_values: Dict[str, MonitorInspection] = {}

    @property
    def monitor_fn(self):
        return self._monitor_fn

    @property
    def mode(self):
        return self._mode

    @property
    def min_delta(self):
        return self._min_delta

    @property
    def patience(self):
        return self._patience

    @property
    def monitor_values(self):
        return self._monitor_values

    def watch(self, source_name, current_monitor_value):
        if source_name not in self._monitor_values.keys():
            self._monitor_values[source_name] = MonitorInspection(value=current_monitor_value)
        else:
            if self._mode is MonitorMode.MIN:
                delta = self._monitor_values[source_name].value - current_monitor_value
            else:
                delta = current_monitor_value - self._monitor_values[source_name].value

            if delta <= self._min_delta:
                self._monitor_values[source_name].with_value(current_monitor_value).reset_inspection_num()
            else:
                self._monitor_values[source_name].add_inspection()
                if self._monitor_values[source_name].inspection_num >= self._patience:
                    raise MonitorPatienceExceeded()



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import ABC, abstractmethod

from visdom import Visdom

from kerosene.loggers.visdom import PlotType
from kerosene.loggers.visdom.data import VisdomData


class VisdomPlot(ABC):

    def __init__(self, visdom: Visdom):
        self._visdom = visdom
        self._window = None

    @property
    def visdom(self):
        return self._visdom

    @property
    def window(self):
        return self._window

    @abstractmethod
    def update(self, visdom_data: VisdomData):
        raise NotImplementedError()


class ImagePlot(VisdomPlot):
    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self.visdom.image(img=visdom_data.y, **visdom_data.params)
        else:
            self.visdom.image(img=visdom_data.y, win=self._window, **visdom_data.params)


class ImagesPlot(VisdomPlot):
    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.images(tensor=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.images(tensor=visdom_data.y, win=self._window, **visdom_data.params)


class LinePlot(VisdomPlot):
    def __init__(self, visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.line(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.line(X=visdom_data.x, Y=visdom_data.y, win=self._window, update='append',
                              name=visdom_data.params['opts']['name'])


class PiePlot(VisdomPlot):
    def __init__(self, visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.pie(X=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.pie(X=visdom_data.y, win=self._window, **visdom_data.params)


class TextPlot(VisdomPlot):
    def __init__(self, visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.text(visdom_data.y)

        else:
            self._visdom.text(visdom_data.y, win=self._window)


class HistogramPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.histogram(X=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.histogram(X=visdom_data.y, win=self._window, **visdom_data.params)


class ScatterPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.scatter(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.scatter(X=visdom_data.x, Y=visdom_data.y, win=self._window, **visdom_data.params)


class StemPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.stem(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.stem(X=visdom_data.x, Y=visdom_data.y, win=self._window, **visdom_data.params)


class HeatmapPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.heatmap(X=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.heatmap(X=visdom_data.y, win=self._window, **visdom_data.params)


class BoxPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.boxplot(X=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.boxplot(X=visdom_data.x, win=self._window, **visdom_data.params)


class SurfacePlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.surf(X=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.surf(X=visdom_data.y, win=self._window, **visdom_data.params)


class ContourPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.contour(X=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.contour(X=visdom_data.y, win=self._window, **visdom_data.params)


class QuiverPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.quiver(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.quiver(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)


class MeshPlot(VisdomPlot):

    def __init__(self, visdom: Visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.mesh(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.mesh(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)


class BarPlot(VisdomPlot):
    def __init__(self, visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.bar(X=visdom_data.x, Y=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.bar(X=visdom_data.x, Y=visdom_data.y, win=self._window, **visdom_data.params)


class MatplotlibPlot(VisdomPlot):
    def __init__(self, visdom):
        super().__init__(visdom)

    def update(self, visdom_data: VisdomData):
        if self._window is None:
            self._window = self._visdom.matplot(plot=visdom_data.y, **visdom_data.params)
        else:
            self._visdom.matplot(plot=visdom_data.y, win=self._window, **visdom_data.params)


class VisdomPlotFactory(object):

    def __init__(self):
        self._plot = {
            PlotType.LINE_PLOT: LinePlot,
            PlotType.IMAGE_PLOT: ImagePlot,
            PlotType.IMAGES_PLOT: ImagesPlot,
            PlotType.PIE_PLOT: PiePlot,
            PlotType.TEXT_PLOT: TextPlot,
            PlotType.BAR_PLOT: BarPlot,
            PlotType.HISTOGRAM_PLOT: HistogramPlot,
            PlotType.SCATTER_PLOT: ScatterPlot,
            PlotType.STEM_PLOT: StemPlot,
            PlotType.HEATMAP_PLOT: HeatmapPlot,
            PlotType.BOX_PLOT: BoxPlot,
            PlotType.SURFACE_PLOT: SurfacePlot,
            PlotType.CONTOUR_PLOT: ContourPlot,
            PlotType.QUIVER_PLOT: QuiverPlot,
            PlotType.MESH_PLOT: MeshPlot,
            PlotType.MATPLOTLIB_PLOT: MatplotlibPlot
        }

    def create_plot(self, visdom, plot_type: PlotType):
        return self._plot[plot_type](visdom)



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from enum import Enum
from typing import Union

import torch
from ignite.metrics import MetricsLambda
from torch import nn
from torch.nn.modules.loss import _Loss, _WeightedLoss

from kerosene.utils.constants import EPSILON
from kerosene.utils.tensors import flatten


class CriterionType(Enum):
    DiceLoss = "DiceLoss"
    GeneralizedDiceLoss = "GeneralizedDiceLoss"
    TverskyLoss = "TverskyLoss"
    FocalTverskyLoss = "FocalTverskyLoss"
    BCELoss = "BCELoss"
    BCEWithLogitsLoss = "BCEWithLogtisLoss"
    PoissonNLLLoss = "PoissonNLLLoss"
    CosineEmbeddingLoss = "CosineEmbeddingLoss"
    CrossEntropyLoss = "CrossEntropyLoss"
    CTCLoss = "CTCLoss"
    HingeEmbeddingLoss = "HingeEmbeddingLoss"
    KLDivLoss = "KLDivLoss"
    L1Loss = "L1Loss"
    MSELoss = "MSELoss"
    MarginRankingLoss = "MarginRankingLoss"
    MultiLabelMarginLoss = "MultiLabelMarginLoss"
    MultiLabelSoftMarginLoss = "MultiLabelSoftMarginLoss"
    MultiMarginLoss = "MultiMarginLoss"
    NLLLoss = "NLLLoss"
    SmoothL1Loss = "SmoothL1Loss"
    SoftMarginLoss = "SoftMarginLoss"
    TripletMarginLoss = "TripletMarginLoss"


class CriterionFactory(object):
    def __init__(self):
        super(CriterionFactory, self).__init__()
        self._criterion = {
            "DiceLoss": DiceLoss,
            "GeneralizedDiceLoss": GeneralizedDiceLoss,
            "TverskyLoss": TverskyLoss,
            "FocalTverskyLoss": FocalTverskyLoss,
            "BCELoss": nn.BCELoss,
            "BCEWithLogitsLoss": nn.BCEWithLogitsLoss,
            "PoissonNLLLoss": nn.PoissonNLLLoss,
            "CosineEmbeddingLoss": nn.CosineEmbeddingLoss,
            "CrossEntropyLoss": nn.CrossEntropyLoss,
            "CTCLoss": nn.CTCLoss,
            "HingeEmbeddingLoss": nn.HingeEmbeddingLoss,
            "KLDivLoss": nn.KLDivLoss,
            "L1Loss": nn.L1Loss,
            "MSELoss": nn.MSELoss,
            "MarginRankingLoss": nn.MarginRankingLoss,
            "MultiLabelMarginLoss": nn.MultiLabelMarginLoss,
            "MultiLabelSoftMarginLoss": nn.MultiLabelSoftMarginLoss,
            "MultiMarginLoss": nn.MultiMarginLoss,
            "NLLLoss": nn.NLLLoss,
            "SmoothL1Loss": nn.SmoothL1Loss,
            "SoftMarginLoss": nn.SoftMarginLoss,
            "TripletMarginLoss": nn.TripletMarginLoss,
        }

    def create(self, criterion_type: Union[str, CriterionType], params):
        return self._criterion[str(criterion_type)](**params) if params is not None else self._criterion[
            str(criterion_type)]()

    def register(self, function: str, creator: _Loss):
        """
        Add a new criterion.
        Args:
           function (str): Criterion's name.
           creator (:obj:`torch.nn.Module`): A torch module object wrapping the new custom criterion function.
        """
        self._criterion[function] = creator


class DiceLoss(_WeightedLoss):
    """
    The Sørensen-Dice Loss.
    """
    SUPPORTED_REDUCTIONS = [None, "mean"]

    def __init__(self, reduction: Union[None, str] = "mean", ignore_index: int = -100, weight: torch.Tensor = None):
        if reduction not in self.SUPPORTED_REDUCTIONS:
            raise NotImplementedError("Reduction type not supported.")
        super(DiceLoss, self).__init__(weight=weight, reduction=reduction)
        self._ignore_index = ignore_index

        if self.weight is not None:
            if self.weight.requires_grad is not False:
                self.weight.requires_grad = False

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        """
        Computes the Sørensen–Dice loss.
        Note that PyTorch optimizers minimize a loss. In this case, we would like to maximize the dice loss so we
        return the negated dice loss.
        Args:
            inputs (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The model prediction on which the loss has to
             be computed.
            targets (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The ground truth.
        Returns:
            :obj:`torch.Tensor`: The Sørensen–Dice loss for each class or reduced according to reduction method.
        """
        if not inputs.size() == targets.size():
            raise ValueError("'Inputs' and 'Targets' must have the same shape.")

        inputs = flatten(inputs)
        targets = flatten(targets).float()

        # Compute per channel Dice Coefficient
        intersection = (inputs * targets).sum(-1)

        if self.weight is not None:
            intersection = self.weight * intersection

        cardinality = (inputs + targets).sum(-1)

        ones = torch.Tensor().new_ones((inputs.size(0),), dtype=torch.float, device=inputs.device)

        dice = ones - (2.0 * intersection / cardinality.clamp(min=EPSILON))

        if self._ignore_index != -100:
            def ignore_index_fn(dice_vector):
                try:
                    indices = list(range(len(dice_vector)))
                    indices.remove(self._ignore_index)
                    return dice_vector[indices]
                except ValueError as e:
                    raise IndexError(
                        "'ignore_index' must be non-negative, and lower than the number of classes in confusion matrix, but {} was given. ".format(
                            self._ignore_index))

            dice = MetricsLambda(ignore_index_fn, dice).compute()

        if self.reduction == "mean":
            dice = dice.mean()

        return dice


class GeneralizedDiceLoss(_Loss):
    """
      The Generalized Sørensen-Dice Loss.
      """
    SUPPORTED_REDUCTIONS = [None, "mean"]

    def __init__(self, reduction: Union[None, str] = "mean", ignore_index: int = -100):
        if reduction not in self.SUPPORTED_REDUCTIONS:
            raise NotImplementedError("Reduction type not supported.")
        super(GeneralizedDiceLoss, self).__init__(reduction=reduction)
        self._ignore_index = ignore_index

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        """
        Computes the Sørensen–Dice loss.
        Note that PyTorch optimizers minimize a loss. In this case, we would like to maximize the dice loss so we
        return the negated dice loss.
        Args:
            inputs (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The model prediction on which the loss has to
             be computed.
            targets (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The ground truth.
        Returns:
            :obj:`torch.Tensor`: The Sørensen–Dice loss for each class or reduced according to reduction method.
        """
        if not inputs.size() == targets.size():
            raise ValueError("'Inputs' and 'Targets' must have the same shape.")

        inputs = flatten(inputs)
        targets = flatten(targets).float()
        ones = torch.Tensor().new_ones((inputs.size(0),), dtype=torch.float, device=inputs.device)
        class_weights = ones / torch.pow(targets.sum(-1), 2).clamp(min=EPSILON)

        # Compute per channel Dice Coefficient
        intersect = (inputs * targets).sum(-1) * class_weights

        denominator = (inputs + targets).sum(-1) * class_weights

        dice = ones - (2.0 * intersect / denominator.clamp(min=EPSILON))

        if self._ignore_index != -100:
            def ignore_index_fn(dice_vector):
                try:
                    indices = list(range(len(dice_vector)))
                    indices.remove(self._ignore_index)
                    return dice_vector[indices]
                except ValueError as e:
                    raise IndexError(
                        "'ignore_index' must be non-negative, and lower than the number of classes in confusion matrix, but {} was given. ".format(
                            self._ignore_index))

            dice = MetricsLambda(ignore_index_fn, dice).compute()

        if self.reduction == "mean":
            dice = dice.mean()

        return dice


class TverskyLoss(_WeightedLoss):
    SUPPORTED_REDUCTIONS = [None, "mean"]

    def __init__(self, reduction: Union[None, str] = "mean", ignore_index: int = -100, weight: torch.Tensor = None,
                 alpha: float = 0.5, beta: float = 0.5):
        if reduction not in self.SUPPORTED_REDUCTIONS:
            raise NotImplementedError("Reduction type not supported.")
        super(TverskyLoss, self).__init__(weight=weight, reduction=reduction)
        self._ignore_index = ignore_index
        self._alpha = alpha
        self._beta = beta

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        """
           Computes the Tversky loss based on https://arxiv.org/pdf/1706.05721.pdf
           Note that PyTorch optimizers minimize a loss. In this case, we would like to maximize the dice loss so we
           return the negated dice loss.
           Args:
               inputs (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The model prediction on which the loss has to
                be computed.
               targets (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The ground truth.
           Returns:
               :obj:`torch.Tensor`: The Tversky loss for each class or reduced according to reduction method.
           """
        if not inputs.size() == targets.size():
            raise ValueError("'Inputs' and 'Targets' must have the same shape.")

        inputs = flatten(inputs)
        targets = flatten(targets).float()
        ones = torch.Tensor().new_ones((inputs.size()), dtype=torch.float, device=inputs.device)

        P_G = (inputs * targets).sum(-1)
        if self.weight is not None:
            P_G = self.weight * P_G

        P_NG = (inputs * (ones - targets)).sum(-1)
        NP_G = ((ones - inputs) * targets).sum(-1)

        ones = torch.Tensor().new_ones((inputs.size(0),), dtype=torch.float, device=inputs.device)
        tversky = P_G / (P_G + self._alpha * P_NG + self._beta * NP_G + EPSILON)

        tversky_loss = ones - tversky

        if self._ignore_index != -100:
            def ignore_index_fn(tversky_vector):
                try:
                    indices = list(range(len(tversky_vector)))
                    indices.remove(self._ignore_index)
                    return tversky_vector[indices]
                except ValueError as e:
                    raise IndexError(
                        "'ignore_index' must be non-negative, and lower than the number of classes in confusion matrix, but {} was given. ".format(
                            self._ignore_index))

            tversky_loss = MetricsLambda(ignore_index_fn, tversky_loss).compute()

        if self.reduction == "mean":
            tversky_loss = tversky_loss.mean()

        return tversky_loss


class FocalTverskyLoss(_WeightedLoss):
    SUPPORTED_REDUCTIONS = [None, "mean"]

    def __init__(self, reduction: Union[None, str] = "mean", ignore_index: int = -100, weight: torch.Tensor = None,
                 alpha: float = 0.5, beta: float = 0.5, gamma: float = 1.0):
        if reduction not in self.SUPPORTED_REDUCTIONS:
            raise NotImplementedError("Reduction type not supported.")
        super(FocalTverskyLoss, self).__init__(weight=weight, reduction=reduction)
        self._ignore_index = ignore_index
        self._alpha = alpha
        self._beta = beta
        self._gamma = gamma

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor):
        """
           Computes the Focal Tversky Loss based on https://arxiv.org/pdf/1708.02002.pdf.
           Note that PyTorch optimizers minimize a loss. In this case, we would like to maximize the dice loss so we
           return the negated dice loss.
           Args:
               inputs (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The model prediction on which the loss has to
                be computed.
               targets (:obj:`torch.Tensor`) : A tensor of shape (B, C, ..). The ground truth.
           Returns:
               :obj:`torch.Tensor`: The Sørensen–Dice loss for each class or reduced according to reduction method.
           """
        if not inputs.size() == targets.size():
            raise ValueError("'Inputs' and 'Targets' must have the same shape.")

        inputs = flatten(inputs)
        targets = flatten(targets).float()
        ones = torch.Tensor().new_ones((inputs.size()), dtype=torch.float, device=inputs.device)

        P_G = (inputs * targets).sum(-1)
        if self.weight is not None:
            P_G = self.weight * P_G

        P_NG = (inputs * (ones - targets)).sum(-1)
        NP_G = ((ones - inputs) * targets).sum(-1)

        ones = torch.Tensor().new_ones((inputs.size(0),), dtype=torch.float, device=inputs.device)
        tversky = P_G / (P_G + self._alpha * P_NG + self._beta * NP_G + EPSILON)

        tversky = ones - (torch.pow((ones - tversky), 1 / self._gamma))

        if self._ignore_index != -100:
            def ignore_index_fn(tversky_vector):
                try:
                    indices = list(range(len(tversky_vector)))
                    indices.remove(self._ignore_index)
                    return tversky_vector[indices]
                except ValueError as e:
                    raise IndexError(
                        "'ignore_index' must be non-negative, and lower than the number of classes in confusion matrix, but {} was given. ".format(
                            self._ignore_index))

            tversky = MetricsLambda(ignore_index_fn, tversky).compute()

        if self.reduction == "mean":
            tversky = tversky.mean()

        return tversky


class WeightedCrossEntropyLoss(_Loss):

    def __init__(self, ignore_index: int = -100):
        super(WeightedCrossEntropyLoss, self).__init__()
        self._ignore_index = ignore_index

    def forward(self, inputs: torch.Tensor, targets: torch.Tensor) -> float:
        """
        Computes the Weighted Cross Entropy Loss (WCE) as described in https://arxiv.org/pdf/1707.03237.pdf
        Args:
            inputs (:obj:`torch.Tensor`): A tensor of shape (B, C, ..). The model's prediction on which the loss has to be computed.
            targets (:obj:`torch.Tensor`): A tensor of shape (B, ..). The ground truth.
       Returns:
           float: the weighted Cross-Entropy loss value.
        """
        class_weights = self.compute_class_weights(inputs)

        return torch.nn.functional.cross_entropy(inputs, targets, weight=class_weights,
                                                 ignore_index=self._ignore_index)

    @staticmethod
    def compute_class_weights(inputs: torch.Tensor):
        """
        Compute weights for each class as described in https://arxiv.org/pdf/1707.03237.pdf
        Args:
            inputs: (:obj:`torch.Tensor`): A tensor of shape (B, C, ..). The model's prediction on which the loss has to be computed.
        Returns:
            :obj:`torch.Tensor`: A tensor containing class weights.
        """
        flattened_inputs = flatten(inputs)
        class_weights = (flattened_inputs.shape[1] - flattened_inputs.sum(-1)) / flattened_inputs.sum(-1)
        return class_weights



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import ABC, abstractmethod


class Gauge(ABC):

    @abstractmethod
    def update(self, **kwargs):
        raise NotImplementedError()

    @abstractmethod
    def compute(self):
        raise NotImplementedError()

    @abstractmethod
    def reset(self):
        raise NotImplementedError()


class AverageGauge(Gauge):

    def __init__(self):
        self._value = 0.0
        self._sum = 0.0
        self._count = 0

    @property
    def count(self):
        return self._count

    def has_been_updated(self):
        return True if self._count > 0 else False

    def update(self, value, n=1):
        self._value = value
        self._sum += value * n
        self._count += n

    def compute(self):
        return self._sum / self._count if self._count != 0 else 0.0

    def reset(self):
        self._value = 0.0
        self._sum = 0.0
        self._count = 0



import unittest

import torch
from hamcrest import *

from kerosene.configs.parsers import CustomYamlParser


class YamlParserTest(unittest.TestCase):

    def setUp(self) -> None:
        self._path = "tests/parsers/test_yaml.yml"
        self._parser = CustomYamlParser()

    def test_should_parse_file_and_create_tensor(self):
        with open(self._path) as file:
            config = self._parser.safe_load(file)

        assert_that(config["weights"], instance_of(torch.Tensor))
        assert_that(config["tuple"], instance_of(tuple))



import unittest

from hamcrest import *

from kerosene.configs.configs import TrainerConfiguration, DatasetConfiguration
from kerosene.utils.configs import configs_to_html


class ConfigTests(unittest.TestCase):

    def setUp(self) -> None:
        self._config_1 = TrainerConfiguration({"num_epochs": 30, "batch_size": 32})
        self._config_2 = DatasetConfiguration({"path": "/home/data", "validation_split": 0.2})

    def test_should_produce_html_from_configs(self):
        config_html = configs_to_html([self._config_1, self._config_2])
        assert_that(config_html[0],
                    is_('<h2>Training Configuration</h2> \n <p>num_epochs: 30</p>\n<p>batch_size: 32</p>'))
        assert_that(config_html[1],
                    is_('<h2>Dataset Configuration</h2> \n <p>path: /home/data</p>\n<p>validation_split: 0.2</p>'))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from abc import abstractmethod
from typing import Dict, Union, List, Optional

import torch
from ignite.metrics import Metric
from torch.utils.data import DataLoader

from kerosene.configs.configs import ModelConfiguration, RunConfiguration
from kerosene.events import BaseEvent, Monitor, Frequency, Phase
from kerosene.events.publishers.base_publisher import EventPublisher
from kerosene.metrics.gauges import AverageGauge
from kerosene.metrics.metrics import MetricFactory, to_tensor
from kerosene.models.models import ModelFactory
from kerosene.nn.apex import ApexModule, ApexLoss
from kerosene.nn.criterions import CriterionFactory
from kerosene.nn.utils.gradients import GradientClippingStrategy, GradientClippingStrategyFactory
from kerosene.optim.optimizers import OptimizerFactory
from kerosene.optim.schedulers import SchedulerFactory
from kerosene.training import Status
from kerosene.training.events import BatchEventPublisherMixin, EpochEventPublisherMixin, \
    TrainingPhaseEventPublisherMixin
from kerosene.utils.devices import on_gpu


class ModelTrainer(ApexModule):
    LOGGER = logging.getLogger("ModelTrainer")

    def __init__(self, model_name, model, criterions, optimizer, scheduler, metric_computers: Dict[str, Metric],
                 gradient_clipping_strategy: GradientClippingStrategy = None):
        super(ModelTrainer, self).__init__(model, optimizer)
        self._model_name = model_name

        self._criterions = criterions
        self._scheduler = scheduler
        self._metric_computers = metric_computers
        self._gradient_clipping_strategy = gradient_clipping_strategy

        self._step_train_loss = {criterion: torch.Tensor().new_zeros((1,)) for criterion in criterions.keys()}
        self._step_valid_loss = {criterion: torch.Tensor().new_zeros((1,)) for criterion in criterions.keys()}
        self._step_test_loss = {criterion: torch.Tensor().new_zeros((1,)) for criterion in criterions.keys()}
        self._step_train_metrics = {metric: torch.Tensor().new_zeros((1,)) for metric in metric_computers.keys()}
        self._step_valid_metrics = {metric: torch.Tensor().new_zeros((1,)) for metric in metric_computers.keys()}
        self._step_test_metrics = {metric: torch.Tensor().new_zeros((1,)) for metric in metric_computers.keys()}

        self._train_loss = {criterion: AverageGauge() for criterion in criterions.keys()}
        self._valid_loss = {criterion: AverageGauge() for criterion in criterions.keys()}
        self._test_loss = {criterion: AverageGauge() for criterion in criterions.keys()}
        self._train_metrics = {metric_name: AverageGauge() for metric_name in metric_computers.keys()}
        self._valid_metrics = {metric_name: AverageGauge() for metric_name in metric_computers.keys()}
        self._test_metrics = {metric_name: AverageGauge() for metric_name in metric_computers.keys()}

        self._status = Status.INITIALIZED

    @property
    def name(self):
        return self._model_name

    @property
    def step_train_loss(self):
        return {loss_name: to_tensor(step_train_loss).cpu() for loss_name, step_train_loss in
                self._step_train_loss.items()}

    @property
    def step_valid_loss(self):
        return {loss_name: to_tensor(step_valid_loss).cpu() for loss_name, step_valid_loss in
                self._step_valid_loss.items()}

    @property
    def step_test_loss(self):
        return {loss_name: to_tensor(step_test_loss).cpu() for loss_name, step_test_loss in
                self._step_test_loss.items()}

    @property
    def step_train_metrics(self):
        return {metric_name: to_tensor(step_train_metric).cpu() for metric_name, step_train_metric in
                self._step_train_metrics.items()}

    @property
    def step_valid_metrics(self):
        return {metric_name: to_tensor(step_valid_metric).cpu() for metric_name, step_valid_metric in
                self._step_valid_metrics.items()}

    @property
    def step_test_metrics(self):
        return {metric_name: to_tensor(step_test_metric).cpu() for metric_name, step_test_metric in
                self._step_test_metrics.items()}

    @property
    def train_loss(self):
        return {loss_name: to_tensor(train_loss.compute()).cpu() for loss_name, train_loss in
                self._train_loss.items()}

    @property
    def valid_loss(self):
        return {loss_name: to_tensor(train_loss.compute()).cpu() for loss_name, train_loss in
                self._valid_loss.items()}

    @property
    def test_loss(self):
        return {loss_name: to_tensor(train_loss.compute()).cpu() for loss_name, train_loss in
                self._test_loss.items()}

    @property
    def train_metrics(self):
        return {metric_name: to_tensor(train_metric.compute()).cpu() for metric_name, train_metric in
                self._train_metrics.items()}

    @property
    def valid_metrics(self):
        return {metric_name: to_tensor(valid_metric.compute()).cpu() for metric_name, valid_metric in
                self._valid_metrics.items()}

    @property
    def test_metrics(self):
        return {metric_name: to_tensor(test_metric.compute()).cpu() for metric_name, test_metric in
                self._test_metrics.items()}

    @property
    def optimizer_state(self):
        return self._optimizer.state_dict()

    @property
    def optimizer_lr(self):
        for param_group in self._optimizer.param_groups:
            return torch.tensor([param_group["lr"]])

    @optimizer_lr.setter
    def optimizer_lr(self, lr):
        for param_group in self._optimizer.param_groups:
            param_group["lr"] = lr

    @property
    def model_state(self):
        return self._model.state_dict()

    @property
    def criterions(self):
        return self._criterions

    @property
    def scheduler(self):
        return self._scheduler

    def step_monitors(self):
        return {Phase.TRAINING: {Monitor.METRICS: self.step_train_metrics,
                                 Monitor.LOSS: {name: value.item() for name, value in self.step_train_loss.items()}},
                Phase.VALIDATION: {Monitor.METRICS: self.step_valid_metrics,
                                   Monitor.LOSS: {name: value.item() for name, value in self.step_valid_loss.items()}},
                Phase.TEST: {Monitor.METRICS: self.step_test_metrics,
                             Monitor.LOSS: {name: value.item() for name, value in self.step_test_loss.items()}}}

    def epoch_monitors(self):
        return {Phase.TRAINING: {Monitor.METRICS: self.train_metrics, Monitor.LOSS: self.train_loss},
                Phase.VALIDATION: {Monitor.METRICS: self.valid_metrics, Monitor.LOSS: self.valid_loss},
                Phase.TEST: {Monitor.METRICS: self.test_metrics, Monitor.LOSS: self.test_loss}}

    def to_state_dict(self, frequency: Frequency = Frequency.EPOCH,
                      training_phase: Union[List[Phase], Phase] = Phase.ALL):
        training_phase = [training_phase] if not isinstance(training_phase, list) else training_phase
        monitors = self.step_monitors() if frequency == Frequency.STEP else self.epoch_monitors()

        return {"Model": self.model_state,
                "Optimizer": self.optimizer_state,
                "Monitors": dict(map(lambda phase: (phase, monitors[phase]), training_phase))}

    def is_active(self) -> bool:
        return self._status is not Status.FINALIZED

    def named_parameters(self, prefix: str = ..., recurse: bool = ...) -> tuple:
        return self.model.named_parameters()

    def forward(self, *input):
        return self._model.forward(*input)

    def train(self, mode=True) -> None:
        self._status = Status.TRAINING
        self._model.train()

    def eval(self) -> None:
        self._status = Status.VALIDATING
        self._model.eval()

    def test(self) -> None:
        self._status = Status.TESTING
        self._model.eval()

    def step(self):
        if self._status is Status.TRAINING:
            if self._should_clip_gradients():
                self._gradient_clipping_strategy.clip(self._model.parameters())
            self._optimizer.step()

    def scheduler_step(self) -> None:
        if self._scheduler is not None:
            self._scheduler.step(sum(list(map(lambda loss: loss.compute(), self._train_loss.values()))))

    def to(self, device: Optional[Union[int, torch.device]] = ..., dtype=..., non_blocking: bool = ...):
        self.model.to(device)

        for criterion in self.criterions.values():
            criterion.to(device)

    def zero_grad(self) -> None:
        self._optimizer.zero_grad()

    def freeze(self) -> None:
        for param in self.parameters():
            param.requires_grad = False

    def unfreeze(self) -> None:
        for param in self.parameters():
            param.requires_grad = True

    def compute_and_update_train_loss(self, name, pred, target) -> Union[ApexLoss, torch.Tensor]:
        self._step_train_loss[name] = self._criterions[name](pred, target)
        self._train_loss[name].update(self._step_train_loss[name].item())

        return ApexLoss(self._amp_id, self._step_train_loss[name], self._optimizer) if self.use_amp else \
            self._step_train_loss[name]

    def compute_and_update_train_losses(self, pred, target) -> Dict[str, Union[ApexLoss, torch.Tensor]]:
        losses = {}
        for name, criterion in self._criterions.items():
            self._step_train_loss[name] = criterion(pred, target)
            self._train_loss[name].update(self._step_train_loss[name].item())
            losses[name] = ApexLoss(self._amp_id, self._step_train_loss[name],
                                    self._optimizer) if self.use_amp else self._step_train_loss[name]
        return losses

    def compute_and_update_valid_loss(self, name, pred, target) -> Union[ApexLoss, torch.Tensor]:
        self._step_valid_loss[name] = self._criterions[name](pred, target)
        self._valid_loss[name].update(self._step_valid_loss[name].item())

        return ApexLoss(self._amp_id, self._step_valid_loss[name], self._optimizer) if self.use_amp else \
            self._step_valid_loss[name]

    def compute_and_update_valid_losses(self, pred, target) -> Dict[str, Union[ApexLoss, torch.Tensor]]:
        losses = {}
        for name, criterion in self._criterions.items():
            self._step_valid_loss[name] = criterion(pred, target)
            self._valid_loss[name].update(self._step_valid_loss[name].item())
            losses[name] = ApexLoss(self._amp_id, self._step_valid_loss[name],
                                    self._optimizer) if self.use_amp else self._step_valid_loss[name]
        return losses

    def compute_and_update_test_loss(self, name, pred, target) -> Union[ApexLoss, torch.Tensor]:
        self._step_test_loss[name] = self._criterions[name](pred, target)
        self._test_loss[name].update(self._step_test_loss[name].item())

        return ApexLoss(self._amp_id, self._step_test_loss[name], self._optimizer) if self.use_amp else \
            self._step_test_loss[name]

    def compute_and_update_test_losses(self, pred, target) -> Dict[str, Union[ApexLoss, torch.Tensor]]:
        losses = {}
        for name, criterion in self._criterions.items():
            self._step_test_loss[name] = criterion(pred, target)
            self._test_loss[name].update(self._step_test_loss[name].item())
            losses[name] = ApexLoss(self._amp_id, self._step_test_loss[name],
                                    self._optimizer) if self.use_amp else self._step_test_loss[name]
        return losses

    def compute_loss(self, name, pred, target) -> Union[ApexLoss, torch.Tensor]:
        loss = self._criterions[name](pred, target)

        return ApexLoss(self._amp_id, loss, self._optimizer) if self.use_amp else loss

    def compute_losses(self, pred, target) -> Dict[str, Union[ApexLoss, torch.Tensor]]:
        losses = {}
        for name, criterion in self._criterions.items():
            loss = criterion(pred, target)
            losses[name] = ApexLoss(self._amp_id, loss, self._optimizer) if self.use_amp else loss
        return losses

    def update_train_loss(self, name, loss) -> None:
        self._step_train_loss[name] = loss if not isinstance(loss, ApexLoss) else loss.loss
        self._train_loss[name].update(self._step_train_loss[name].item())

    def update_train_losses(self, losses) -> None:
        for name, value in losses.items():
            self._step_train_loss[name] = value if not isinstance(value, ApexLoss) else value.loss
            self._train_loss[name].update(self._step_train_loss[name].item())

    def update_valid_loss(self, name, loss) -> None:
        self._step_valid_loss[name] = loss if not isinstance(loss, ApexLoss) else loss.loss
        self._valid_loss[name].update(self._step_valid_loss[name].item())

    def update_valid_losses(self, losses) -> None:
        for name, value in losses.items():
            self._step_valid_loss[name] = value if not isinstance(value, ApexLoss) else value.loss
            self._valid_loss[name].update(self._step_valid_loss[name].item())

    def update_test_loss(self, name, loss) -> None:
        self._step_test_loss[name] = loss if not isinstance(loss, ApexLoss) else loss.loss
        self._test_loss[name].update(self._step_test_loss[name].item())

    def update_test_losses(self, losses) -> None:
        for name, value in losses.items():
            self._step_test_loss[name] = value if not isinstance(value, ApexLoss) else value.loss
            self._test_loss[name].update(self._step_test_loss[name].item())

    def compute_metrics(self, pred, target):
        metrics = dict()
        for metric_name, metric_computer in self._metric_computers.items():
            metric_computer.update((pred, target))
            metric = metric_computer.compute()
            metric_computer.reset()
            metrics[metric_name] = metric

        return metrics

    def update_train_metrics(self, metrics: dict):
        self._step_train_metrics = metrics
        for train_metric, step_train_metric in zip(list(self._train_metrics.values()),
                                                   list(self._step_train_metrics.values())):
            train_metric.update(step_train_metric)

    def update_valid_metrics(self, metric: dict):
        self._step_valid_metrics = metric
        for valid_metric, step_valid_metric in zip(list(self._valid_metrics.values()),
                                                   list(self._step_valid_metrics.values())):
            valid_metric.update(step_valid_metric)

    def update_test_metrics(self, metric: dict):
        self._step_test_metrics = metric
        for test_metric, step_test_metric in zip(list(self._test_metrics.values()),
                                                 list(self._step_test_metrics.values())):
            test_metric.update(step_test_metric)

    def reset(self):
        [metric_computer.reset() for metric_computer in self._metric_computers.values()]
        [metric.reset() for metric_name, metric in self._train_metrics.items()]
        [metric.reset() for metric_name, metric in self._valid_metrics.items()]
        [metric.reset() for metric_name, metric in self._test_metrics.items()]
        [loss.reset() for loss_name, loss in self._train_loss.items()]
        [loss.reset() for loss_name, loss in self._valid_loss.items()]
        [loss.reset() for loss_name, loss in self._test_loss.items()]

    def finalize(self) -> None:
        self._status = Status.FINALIZED

    def _should_clip_gradients(self):
        return self._gradient_clipping_strategy is not None


class Trainer(BatchEventPublisherMixin, EpochEventPublisherMixin, TrainingPhaseEventPublisherMixin, EventPublisher):
    LOGGER = logging.getLogger("Trainer")

    def __init__(self, name, train_data_loader: DataLoader, valid_data_loader: DataLoader,
                 test_data_loader: Union[DataLoader, None], model_trainers: Union[List[ModelTrainer], ModelTrainer],
                 run_config: RunConfiguration):
        super().__init__()
        self._name = name

        self._train_data_loader = train_data_loader
        self._valid_data_loader = valid_data_loader
        self._test_data_loader = test_data_loader

        self._model_trainers = model_trainers if isinstance(model_trainers, list) else [model_trainers]
        self._device = run_config.device

        self._current_train_batch = 0
        self._current_valid_batch = 0
        self._current_test_batch = 0
        self._current_epoch = 0

        self._custom_variables = {}

        for amp_id, model_trainer in enumerate(self._model_trainers):
            if on_gpu(self._device):
                model_trainer.to(self._device)
            model_trainer.initialize(amp_id, len(self._model_trainers), run_config.use_amp, run_config.amp_opt_level,
                                     self._device)

    @property
    def name(self):
        return self._name

    @property
    def train_data_loader(self):
        return self._train_data_loader

    @property
    def valid_data_loader(self):
        return self._valid_data_loader

    @property
    def test_data_loader(self):
        return self._test_data_loader

    @property
    def model_trainers(self):
        return self._model_trainers

    @property
    def current_train_step(self):
        return self._current_epoch * len(self._train_data_loader) + self._current_train_batch

    @property
    def current_valid_step(self):
        return self._current_epoch * len(self._valid_data_loader) + self._current_valid_batch

    @property
    def current_test_step(self):
        return self._current_epoch * len(
            self._test_data_loader) + self._current_test_batch if self._test_data_loader is not None else 0

    @property
    def epoch(self):
        return self._current_epoch

    @property
    def custom_variables(self):
        return self._custom_variables

    @property
    def nb_of_train_batch(self):
        return len(self._train_data_loader)

    @property
    def nb_of_valid_batch(self):
        return len(self._valid_data_loader)

    @property
    def nb_on_test_batch(self):
        return len(self._test_data_loader)

    @property
    def sender(self):
        return self

    @property
    def status(self):
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

    def step_monitors(self, phase: Phase):
        if phase != Phase.ALL:
            monitors = dict(
                map(lambda model: (model.name, {phase: model.step_monitors()[phase]}), self._model_trainers))
        else:
            monitors = dict(
                map(lambda model: (model.name, model.step_monitors()), self._model_trainers))

        return monitors

    def epoch_monitors(self, phase: Phase):
        if phase != Phase.ALL:
            monitors = dict(
                map(lambda model: (model.name, {phase: model.epoch_monitors()[phase]}), self._model_trainers))
        else:
            monitors = dict(
                map(lambda model: (model.name, model.epoch_monitors()), self._model_trainers))

        return monitors

    @abstractmethod
    def train_step(self, inputs, target):
        raise NotImplementedError()

    @abstractmethod
    def validate_step(self, inputs, target):
        raise NotImplementedError()

    @abstractmethod
    def test_step(self, inputs, target):
        raise NotImplementedError()

    @abstractmethod
    def scheduler_step(self):
        raise NotImplementedError()

    def is_active(self):
        return self._status is not Status.FINALIZED

    def train(self, nb_epoch):
        self._on_training_begin()

        for self._current_epoch in range(0, nb_epoch):
            if self._at_least_one_model_is_active():
                self._on_training_begin()
                self._on_epoch_begin()
                self._on_train_epoch_begin()
                self._train_epoch()
                self._on_train_epoch_end()
                self._on_epoch_end()
                self._on_training_end()
                self._on_valid_begin()
                self._on_epoch_begin()
                self._on_valid_epoch_begin()
                self._validate_epoch()
                self._on_valid_epoch_end()
                self._on_epoch_end()
                self._on_valid_end()
                self._on_test_begin()
                self._on_epoch_begin()
                self._on_test_epoch_begin()
                self._test_epoch()
                self._on_test_epoch_end()
                self._on_epoch_end()
                self._on_test_end()
            else:
                self._finalize()

    def with_event_handler(self, handler, event: BaseEvent):
        if event in self._event_handlers.keys():
            self._event_handlers[event].append(handler)
        else:
            self._event_handlers[event] = [handler]

        return self

    def _train_epoch(self):
        for model_trainer in self._model_trainers:
            model_trainer.train()

        for self._current_train_batch, (inputs, target) in enumerate(self._train_data_loader):
            self._on_batch_begin()
            self._on_train_batch_begin()

            inputs = [single_input.to(self._device, non_blocking=True) for single_input in
                      inputs] if isinstance(inputs, list) else inputs.to(self._device, non_blocking=True)

            target = [single_target.to(self._device, non_blocking=True) for single_target in
                      target] if isinstance(target, list) else target.to(self._device, non_blocking=True)

            self.train_step(inputs, target)
            self._on_train_batch_end()
            self._on_batch_end()

    def _validate_epoch(self):
        for model_trainer in self._model_trainers:
            model_trainer.eval()

        with torch.no_grad():
            for self._current_valid_batch, (inputs, target) in enumerate(self._valid_data_loader):
                self._on_valid_batch_begin()

                inputs = [single_input.to(self._device, non_blocking=True) for single_input in
                          inputs] if isinstance(inputs, list) else inputs.to(self._device, non_blocking=True)

                target = [single_target.to(self._device, non_blocking=True) for single_target in
                          target] if isinstance(target, list) else target.to(self._device, non_blocking=True)

                self.validate_step(inputs, target)
                self._on_valid_batch_end()
                self._on_batch_end()

    def _test_epoch(self):
        if self._test_data_loader is not None:
            for model_trainer in self._model_trainers:
                model_trainer.eval()

            with torch.no_grad():
                for self._current_test_batch, (inputs, target) in enumerate(self._test_data_loader):
                    self._on_test_batch_begin()

                    inputs = [single_input.to(self._device, non_blocking=True) for single_input in
                              inputs] if isinstance(inputs, list) else inputs.to(self._device, non_blocking=True)

                    target = [single_target.to(self._device, non_blocking=True) for single_target in
                              target] if isinstance(target, list) else target.to(self._device, non_blocking=True)

                    self.test_step(inputs, target)
                    self._on_test_batch_end()
                    self._on_batch_end()

    def _reset_model_trainers(self):
        for model_trainer in self._model_trainers:
            model_trainer.reset()

    def _at_least_one_model_is_active(self):
        return any(model_trainer.is_active for model_trainer in self._model_trainers)


class SimpleTrainer(Trainer):

    def train_step(self, inputs, target):
        model = self._model_trainers[0]

        pred = model.forward(inputs)
        metric = model.compute_metrics(pred, target)
        model.update_train_metrics(metric)
        losses = model.compute_and_update_train_losses(pred, target)

        model.zero_grad()
        for loss in losses.values():
            loss.backward()
        model.step()

    def validate_step(self, inputs, target):
        model = self._model_trainers[0]

        pred = model.forward(inputs)
        metric = model.compute_metrics(pred, target)
        model.update_valid_metrics(metric)
        model.compute_and_update_valid_losses(pred, target)

    def test_step(self, inputs, target):
        model = self._model_trainers[0]

        pred = model.forward(inputs)
        metric = model.compute_metrics(pred, target)
        model.update_test_metrics(metric)
        model.compute_and_update_test_losses(pred, target)

    def scheduler_step(self):
        self._model_trainers[0].scheduler_step()


class ModelTrainerFactory(object):
    def __init__(self, model=None, model_factory: ModelFactory = None, optimizer_factory=OptimizerFactory(),
                 scheduler_factory=SchedulerFactory(), criterion_factory=CriterionFactory(),
                 metric_factory=MetricFactory(), gradient_clipping_strategy_factory=GradientClippingStrategyFactory()):
        self._model = model
        self._model_factory = model_factory
        self._optimizer_factory = optimizer_factory
        self._scheduler_factory = scheduler_factory
        self._criterion_factory = criterion_factory
        self._metric_factory = metric_factory
        self._gradient_clipping_strategy_factory = gradient_clipping_strategy_factory

        assert (self._model is not None) or (
                self._model_factory is not None), "A model or a model factory must be provided !"

    def create(self, model_trainer_configs: Union[List[ModelConfiguration], ModelConfiguration]):
        model_trainer_configs = [model_trainer_configs] if isinstance(model_trainer_configs,
                                                                      ModelConfiguration) else model_trainer_configs
        model_trainers = list(
            map(lambda model_trainer_config: self._create(model_trainer_config), model_trainer_configs))

        return model_trainers if len(model_trainers) > 1 else model_trainers[0]

    def _create(self, model_config: ModelConfiguration):
        model = self._model if self._model is not None else self._model_factory.create(model_config.type,
                                                                                       model_config.params)
        optimizer = self._optimizer_factory.create(model_config.optimizer_type,
                                                   model.parameters(),
                                                   model_config.optimizer_params)
        scheduler = self._scheduler_factory.create(model_config.scheduler_type, optimizer,
                                                   model_config.scheduler_params)

        criterions = {
            criterion_config.name: self._criterion_factory.create(criterion_config.type, criterion_config.params) for
            criterion_config in model_config.criterions_configs}

        metrics = {
            metric_config.name: self._metric_factory.create(metric_config.type, metric_config.params) for
            metric_config in model_config.metrics_configs}

        gradient_clipping_strategy = self._gradient_clipping_strategy_factory.create(
            model_config.gradient_clipping_strategy, model_config.gradient_clipping_params)

        return ModelTrainer(model_config.name, model, criterions, optimizer, scheduler, metrics,
                            gradient_clipping_strategy)



import os
import unittest

from hamcrest import *
from ignite.metrics import Accuracy, Metric, Recall
from mockito import mock, verify, spy
from torch import nn
from torch.nn.modules.loss import _Loss
from torch.optim import SGD
from torch.optim import lr_scheduler

from kerosene.configs.configs import ModelConfiguration, RunConfiguration
from kerosene.configs.parsers import YamlConfigurationParser
from kerosene.nn.utils.gradients import GradientClippingStrategy
from kerosene.training.trainers import ModelTrainer, ModelTrainerFactory
from tests.constants import *
from tests.functionals.models import SimpleNet


# noinspection PyUnresolvedReferences
class ModelTrainerTest(unittest.TestCase):
    MODEL_NAME = "Harry Potter"
    LOSS_NAME = "CrossEntropy"

    def setUp(self) -> None:
        self._model = SimpleNet()
        self._criterion_mock = spy(nn.CrossEntropyLoss())
        self._optimizer_mock = spy(SGD(self._model.parameters(), lr=0.001))
        self._scheduler_mock = mock(lr_scheduler)
        self._accuracy_computer_mock = spy(Accuracy())
        self._recall_computer_mock = spy(Recall())
        self._gradient_clipping_strategy = None

        self._model_trainer = ModelTrainer(self.MODEL_NAME, self._model, {self.LOSS_NAME: self._criterion_mock},
                                           self._optimizer_mock, self._scheduler_mock,
                                           {"Accuracy": self._accuracy_computer_mock,
                                            "Recall": self._recall_computer_mock}, self._gradient_clipping_strategy)

        self._gradient_clipping_strategy = mock(GradientClippingStrategy)

    def tearDown(self) -> None:
        super().tearDown()

    def test_should_compute_train_metric_and_update_state(self):
        metric = self._model_trainer.compute_metrics(MODEL_PREDICTION_CLASS_0, TARGET_CLASS_0)
        self._model_trainer.update_train_metrics(metric)
        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ONE))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ONE))
        verify(self._accuracy_computer_mock).update((MODEL_PREDICTION_CLASS_0, TARGET_CLASS_0))

        metric = self._model_trainer.compute_metrics(MODEL_PREDICTION_CLASS_1, TARGET_CLASS_0)
        self._model_trainer.update_train_metrics(metric)
        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ONE / 2))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ZERO))

        verify(self._accuracy_computer_mock, times=2).compute()

        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_test_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.test_loss, equal_to({self.LOSS_NAME: ZERO}))

    def test_should_compute_valid_metric_and_update_state(self):
        metric = self._model_trainer.compute_metrics(MODEL_PREDICTION_CLASS_0, TARGET_CLASS_0)
        self._model_trainer.update_valid_metrics(metric)
        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ONE))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ONE))
        verify(self._accuracy_computer_mock).update((MODEL_PREDICTION_CLASS_0, TARGET_CLASS_0))

        metric = self._model_trainer.compute_metrics(MODEL_PREDICTION_CLASS_1, TARGET_CLASS_0)
        self._model_trainer.update_valid_metrics(metric)
        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ONE / 2))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ZERO))

        verify(self._accuracy_computer_mock, times=2).compute()
        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_test_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.test_loss, equal_to({self.LOSS_NAME: ZERO}))

    def test_should_compute_test_metric_and_update_state(self):
        metric = self._model_trainer.compute_metrics(MODEL_PREDICTION_CLASS_0, TARGET_CLASS_0)
        self._model_trainer.update_test_metrics(metric)
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ONE))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ONE))
        verify(self._accuracy_computer_mock).update((MODEL_PREDICTION_CLASS_0, TARGET_CLASS_0))

        metric = self._model_trainer.compute_metrics(MODEL_PREDICTION_CLASS_1, TARGET_CLASS_0)
        self._model_trainer.update_test_metrics(metric)
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ONE / 2))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ZERO))

        verify(self._accuracy_computer_mock, times=2).compute()

        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_test_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.test_loss, equal_to({self.LOSS_NAME: ZERO}))

    def test_should_compute_train_loss_and_update_state(self):
        loss = self._model_trainer.compute_and_update_train_loss(self.LOSS_NAME, MODEL_PREDICTION_CLASS_0,
                                                                 TARGET_CLASS_0)

        assert_that(loss._loss, close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.step_train_loss[self.LOSS_NAME],
                    close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.train_loss[self.LOSS_NAME], close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))

        loss = self._model_trainer.compute_and_update_train_loss(self.LOSS_NAME, MODEL_PREDICTION_CLASS_0,
                                                                 TARGET_CLASS_1)
        assert_that(loss._loss, close_to(MAXIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.step_train_loss[self.LOSS_NAME],
                    close_to(MAXIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.train_loss[self.LOSS_NAME], close_to(AVERAGED_BINARY_CROSS_ENTROPY_LOSS, DELTA))

        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_test_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.test_loss, equal_to({self.LOSS_NAME: ZERO}))

    def test_should_compute_valid_loss_and_update_state(self):
        loss = self._model_trainer.compute_and_update_valid_loss(self.LOSS_NAME, MODEL_PREDICTION_CLASS_0,
                                                                 TARGET_CLASS_0)

        assert_that(loss._loss, close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.step_valid_loss[self.LOSS_NAME],
                    close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.valid_loss[self.LOSS_NAME], close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))

        loss = self._model_trainer.compute_and_update_valid_loss(self.LOSS_NAME, MODEL_PREDICTION_CLASS_0,
                                                                 TARGET_CLASS_1)
        assert_that(loss._loss, close_to(MAXIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.step_valid_loss[self.LOSS_NAME],
                    close_to(MAXIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.valid_loss[self.LOSS_NAME], close_to(AVERAGED_BINARY_CROSS_ENTROPY_LOSS, DELTA))

        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ZERO))

        assert_that(self._model_trainer.step_train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_test_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.test_loss, equal_to({self.LOSS_NAME: ZERO}))

    def test_should_compute_test_loss_and_update_state(self):
        loss = self._model_trainer.compute_and_update_test_loss(self.LOSS_NAME, MODEL_PREDICTION_CLASS_0,
                                                                TARGET_CLASS_0)

        assert_that(loss._loss, close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.step_test_loss[self.LOSS_NAME],
                    close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.test_loss[self.LOSS_NAME], close_to(MINIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))

        loss = self._model_trainer.compute_and_update_test_losses(MODEL_PREDICTION_CLASS_0, TARGET_CLASS_1)
        assert_that(loss[self.LOSS_NAME]._loss, close_to(MAXIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.step_test_loss[self.LOSS_NAME],
                    close_to(MAXIMUM_BINARY_CROSS_ENTROPY_LOSS, DELTA))
        assert_that(self._model_trainer.test_loss[self.LOSS_NAME], close_to(AVERAGED_BINARY_CROSS_ENTROPY_LOSS, DELTA))

        assert_that(self._model_trainer.train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_valid_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_test_metrics["Accuracy"], equal_to(ZERO))
        assert_that(self._model_trainer.step_train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.step_valid_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.train_loss, equal_to({self.LOSS_NAME: ZERO}))
        assert_that(self._model_trainer.valid_loss, equal_to({self.LOSS_NAME: ZERO}))


class ModelTrainerFactoryTest(unittest.TestCase):
    VALID_CONFIG_FILE_PATH = "tests/configs/valid_config.yml"
    MODELS_CONFIG_YML_TAG = "models"
    SIMPLE_NET_NAME = "SimpleNet"

    @classmethod
    def setUpClass(cls) -> None:
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"

    def setUp(self) -> None:
        config_dict = YamlConfigurationParser.parse_section(self.VALID_CONFIG_FILE_PATH, self.MODELS_CONFIG_YML_TAG)
        self._model_trainer_config = ModelConfiguration.from_dict(self.SIMPLE_NET_NAME,
                                                                  config_dict[self.SIMPLE_NET_NAME])
        self._run_config = RunConfiguration()
        self._model = nn.Conv2d(1, 32, (3, 3))

    def test_should_create_model_trainer_with_config(self):
        model_trainer = ModelTrainerFactory(model=self._model).create(self._model_trainer_config)
        assert_that(model_trainer, instance_of(ModelTrainer))
        assert_that(model_trainer.name, is_(self.SIMPLE_NET_NAME))
        assert_that(len(model_trainer.criterions), is_(2))
        [assert_that(model_trainer.criterions[criterion], instance_of(_Loss)) for criterion in
         model_trainer.criterions.keys()]
        assert_that(model_trainer.optimizer, instance_of(torch.optim.SGD))
        assert_that(model_trainer.scheduler, instance_of(torch.optim.lr_scheduler.ReduceLROnPlateau))
        assert_that(len(model_trainer._metric_computers.keys()), is_(2))
        [assert_that(model_trainer._metric_computers[metric], instance_of(Metric)) for metric in
         model_trainer._metric_computers.keys()]



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import ABC, abstractmethod

from kerosene.events import BaseEvent, TemporalEvent


class EventPublisher(ABC):
    def __init__(self):
        self._event_handlers = {}

    @property
    @abstractmethod
    def sender(self):
        raise NotImplementedError()

    @abstractmethod
    def with_event_handler(self, handler, event: BaseEvent):
        raise NotImplementedError()

    def fire(self, temporal_event: TemporalEvent, monitors: dict = None):
        if temporal_event.event in self._event_handlers.keys():
            for handler in self._event_handlers[temporal_event.event]:
                handler(temporal_event, monitors, self.sender)



from enum import Enum


class PlotType(Enum):
    LINE_PLOT = "Line Plot"
    IMAGES_PLOT = "Images Plot"
    IMAGE_PLOT = "Image Plot"
    PIE_PLOT = "Pie Plot"
    TEXT_PLOT = "Text Plot"
    HISTOGRAM_PLOT = "Histogram Plot"
    SCATTER_PLOT = "Scatter Plot"
    STEM_PLOT = "Stem Plot"
    HEATMAP_PLOT = "Heatmap Plot"
    BAR_PLOT = "Bar Plot"
    BOX_PLOT = "Box Plot"
    SURFACE_PLOT = "Surface Plot"
    CONTOUR_PLOT = "Contour Plot"
    QUIVER_PLOT = "Quiver Plot"
    MESH_PLOT = "Mesh Plot"
    MATPLOTLIB_PLOT = "Matplotlib Plot"

    def __str__(self):
        return self.value


class PlotFrequency(Enum):
    EVERY_STEP = "Step"
    EVERY_EPOCH = "Epoch"

    def __str__(self):
        return self.value



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from abc import ABC
from enum import Enum
from typing import Dict, Union, List

from kerosene.events import Phase, TemporalEvent, BaseEvent
from kerosene.events.handlers.base_handler import EventHandler
from kerosene.training import Status, BaseStatus
from kerosene.training.events import Event
from kerosene.training.trainers import Trainer


class ConsoleColors(Enum):
    PURPLE = '\033[95m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    YELLOW = '\033[93m'
    RED = '\033[91m'
    ENDC = '\033[0m'
    NONE = ""

    def __str__(self):
        return self.value


class StatusConsoleColorPalette(Dict[Status, ConsoleColors]):
    DEFAULT = {
        Phase.TRAINING: ConsoleColors.GREEN,
        Phase.VALIDATION: ConsoleColors.BLUE,
        Phase.TEST: ConsoleColors.PURPLE
    }


class BaseConsoleLogger(EventHandler, ABC):
    LOGGER = logging.getLogger("ConsoleLogger")

    def __init__(self, supported_events: List[Union[BaseEvent, TemporalEvent]] = None, every=1):
        super(BaseConsoleLogger, self).__init__(supported_events, every)


class ColoredConsoleLogger(BaseConsoleLogger, ABC):

    def __init__(self, supported_events: List[Union[BaseEvent, TemporalEvent]] = None, every=1,
                 colors: Dict[object, ConsoleColors] = None):
        super().__init__(supported_events, every)
        self._colors = colors if colors is not None else {}

    def color(self, text, color_key):
        return str(self._colors.get(color_key, ConsoleColors.NONE)) + str(text) + str(ConsoleColors.ENDC)


class PrintTrainingStatus(ColoredConsoleLogger):
    SUPPORTED_EVENTS = [Event.ON_BATCH_END, Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END,
                        Event.ON_TEST_BATCH_END]

    def __init__(self, every=1, colors: Union[Dict[BaseStatus, ConsoleColors], StatusConsoleColorPalette] = None):
        super().__init__(self.SUPPORTED_EVENTS, every, colors)

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        if self.should_handle(event):
            self.print_status(event.phase, trainer.epoch, trainer.current_train_step, trainer.current_valid_step,
                              trainer.current_test_step)

    def print_status(self, status, epoch, train_step, valid_step, test_step):
        self.LOGGER.info(
            "\nCurrent states: {} |  Epoch: {} | Training step: {} | Validation step: {} | Test step: {}\n".format(
                self.color(status, color_key=status), epoch, train_step, valid_step, test_step))


class PrintMonitors(BaseConsoleLogger):
    SUPPORTED_EVENTS = [Event.ON_BATCH_END, Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END,
                        Event.ON_TEST_BATCH_END]

    def __init__(self, every=1):
        super().__init__(self.SUPPORTED_EVENTS, every)
        self._monitors = {}

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        if self.should_handle(event):
            for model, monitor in monitors.items():
                if model in self._monitors:
                    self._monitors[model].update(monitor)
                else:
                    self._monitors[model] = monitor

                self.LOGGER.info("Model {}, {}".format(model, self._monitors[model]))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os
from glob import glob
from typing import List, Tuple


def split_filename(filepath: str) -> Tuple[str, str, str]:
    """
    Split a filepath into the directory, base, and extension

    Args:
        filepath (str): The base file path.

    Returns:
        Tuple: The complete file path, base path and file extension.
    """
    path = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    base, ext = os.path.splitext(filename)
    if ext == '.gz':
        base, ext2 = os.path.splitext(base)
        ext = ext2 + ext
    return path, base, ext


def extract_file_paths(path: str, ext='*.nii*') -> List[str]:
    """
    Grab all `ext` files in a directory and sort them for consistency.

    Args:
        path (str): File path.
        ext (str): File's extension to grab.

    Returns:
        list: A list of string containing every file paths.
    """
    file_paths = sorted(glob(os.path.join(path, ext)))
    return file_paths


def should_create_dir(path, dir_name):
    return not os.path.exists(os.path.join(path, dir_name))



import logging
import os
import sys
from argparse import ArgumentParser

from kerosene.configs.configs import RunConfiguration

sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)) + '/../../../')
import torchvision
from torchvision.transforms import Compose, ToTensor, Normalize

from kerosene.configs.parsers import YamlConfigurationParser
from kerosene.events import Event
from kerosene.dataloaders.factories import DataloaderFactory
from kerosene.events.handlers.console import ConsoleLogger
from kerosene.events.handlers.visdom.config import VisdomConfiguration
from kerosene.events.handlers.visdom.visdom import VisdomLogger
from kerosene.events.preprocessors.visdom import PlotAllModelStateVariables
from kerosene.training.trainers import ModelTrainerFactory
from tests.functionals.models import SimpleNet
from tests.functionals.distributed.mnist_trainer import MNISTTrainer


class ArgsParserFactory(object):

    @staticmethod
    def create_parser():
        parser = ArgumentParser(description='DeepNormalize Training')
        parser.add_argument("--use_amp", dest="use_amp", action="store_true", default=True)
        parser.add_argument("--amp-opt-level", dest="amp_opt_level", type=str, default="O2",
                            help="O0 - FP32 training, O1 - Mixed Precision (recommended), O2 - Almost FP16 Mixed Precision, O3 - FP16 Training.")
        parser.add_argument("--local_rank", dest="local_rank", default=0, type=int, help="The local_rank of the GPU.")
        return parser


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    CONFIG_FILE_PATH = "config.yml"
    args = ArgsParserFactory.create_parser().parse_args()
    run_config = RunConfiguration(args.use_amp, args.amp_opt_level, args.local_rank)

    model_trainer_config, training_config = YamlConfigurationParser.parse(CONFIG_FILE_PATH)

    # Initialize the dataset. This is the only part the user must define manually.
    train_dataset = torchvision.datasets.MNIST('./files/', train=True, download=True, transform=Compose(
        [ToTensor(), Normalize((0.1307,), (0.3081,))]))
    test_dataset = torchvision.datasets.MNIST('./files/', train=False, download=True, transform=Compose(
        [ToTensor(), Normalize((0.1307,), (0.3081,))]))

    # Initialize loaders.
    train_loader, valid_loader = DataloaderFactory(train_dataset, test_dataset).create(run_config, training_config)

    # Initialize the loggers.
    if run_config.local_rank == 0:
        visdom_logger = VisdomLogger(VisdomConfiguration.from_yml(CONFIG_FILE_PATH))

    # Initialize the model trainers.
    model_trainer = ModelTrainerFactory(model=SimpleNet()).create(model_trainer_config, run_config)

    # Train with the training strategy.
    if run_config.local_rank == 0:
        trainer = MNISTTrainer(training_config, model_trainer, train_loader, valid_loader, run_config) \
            .with_event_handler(ConsoleLogger(), Event.ON_EPOCH_END) \
            .with_event_handler(visdom_logger, Event.ON_EPOCH_END, PlotAllModelStateVariables()) \
            .train(training_config.nb_epochs)
    else:
        trainer = MNISTTrainer(training_config, model_trainer, train_loader, valid_loader, run_config) \
            .train(training_config.nb_epochs)



import logging
from abc import ABC, abstractmethod
from typing import Union, List, Dict

from visdom import Visdom

from kerosene.loggers.visdom.config import VisdomConfiguration
from kerosene.loggers.visdom.data import VisdomData
from kerosene.loggers.visdom.plots import VisdomPlotFactory, VisdomPlot


class BaseVisdomLogger(ABC):

    def __init__(self, visdom_config: VisdomConfiguration):
        self._visdom_config = visdom_config

    @abstractmethod
    def __call__(self, visdom_data: Union[List[VisdomData], VisdomData]):
        raise NotImplementedError()

    @property
    def visdom_config(self):
        return self._visdom_config


class VisdomLogger(BaseVisdomLogger):
    LOGGER = logging.getLogger("VisdomLogger")

    def __init__(self, visdom_config: VisdomConfiguration, plot_factory: VisdomPlotFactory = VisdomPlotFactory()):
        super().__init__(visdom_config)
        self._plots: Dict[int, VisdomPlot] = {}
        self._plot_factory = plot_factory

        self._visdom = Visdom(server=visdom_config.server, port=visdom_config.port, env=visdom_config.env)

    @property
    def plots(self):
        return self._plots

    def __call__(self, visdom_data: Union[List[VisdomData], VisdomData]):
        visdom_data = [visdom_data] if not hasattr(visdom_data, '__iter__') else visdom_data

        for visdom_datum in visdom_data:
            if visdom_datum.id not in self._plots.keys():
                self._plots[visdom_datum.id] = self._plot_factory.create_plot(self._visdom, visdom_datum.plot_type)

            self._plots[visdom_datum.id].update(visdom_datum)


class DummyVisdomLogger(BaseVisdomLogger):
    LOGGER = logging.getLogger("DummyVisdomLogger")

    def __init__(self, visdom_config: VisdomConfiguration = None):
        super().__init__(visdom_config)

    def __call__(self, visdom_data: Union[List[VisdomData], VisdomData]):
        if not isinstance(visdom_data, list):
            visdom_data = [visdom_data]

        self.LOGGER.debug("\n".join(list(map(lambda visdom_datum: str(visdom_datum), visdom_data))))



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from abc import ABC
from typing import Union, List

from kerosene.events import TemporalEvent, Monitor, BaseEvent
from kerosene.events.handlers.base_handler import EventHandler
from kerosene.loggers.visdom import PlotType
from kerosene.loggers.visdom.visdom import VisdomLogger, VisdomData
from kerosene.training.events import Event
from kerosene.training.trainers import Trainer, ModelTrainer


class BaseVisdomHandler(EventHandler, ABC):
    def __init__(self, supported_events: List[Union[BaseEvent, TemporalEvent]], visdom_logger: VisdomLogger, every=1):
        super().__init__(supported_events, every)
        self._visdom_logger = visdom_logger

    @property
    def visdom_logger(self):
        return self._visdom_logger

    def create_visdom_data(self, *args, **kwargs):
        raise NotImplementedError()

    def flatten(self, list_of_visdom_data):
        return [item for sublist in list_of_visdom_data for item in sublist]


class PlotMonitors(BaseVisdomHandler):
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END, Event.ON_TEST_BATCH_END,
                        Event.ON_BATCH_END]

    def __init__(self, visdom_logger: VisdomLogger, every=1):
        super().__init__(self.SUPPORTED_EVENTS, visdom_logger, every)
        self._plot_losses = PlotLosses(visdom_logger, every)
        self._plot_metrics = PlotMetrics(visdom_logger, every)

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        if self.should_handle(event):
            self._plot_losses(event, monitors, trainer)
            self._plot_metrics(event, monitors, trainer)

    def create_visdom_data(self, event, model_name, monitors):
        data = self._plot_losses.create_visdom_data(event, model_name, monitors)
        data.extend(self._plot_metrics.create_visdom_data(event, model_name, monitors))


class PlotLosses(BaseVisdomHandler):
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END, Event.ON_TEST_BATCH_END,
                        Event.ON_BATCH_END]

    def __init__(self, visdom_logger: VisdomLogger, every=1):
        super().__init__(self.SUPPORTED_EVENTS, visdom_logger, every)

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        data = list()

        if self.should_handle(event):
            for model_name, monitor in monitors.items():
                data.extend(self.create_visdom_data(event, model_name, monitor[event.phase][Monitor.LOSS]))

        if data is not None:
            self.visdom_logger(data)

    def create_visdom_data(self, event, model_name, monitors):
        return [VisdomData(model_name, loss_name, PlotType.LINE_PLOT, event.frequency, [[event.iteration]],
                           [[loss_value]], params={'opts': {'xlabel': str(event.frequency), 'ylabel': loss_name,
                                                            'title': "{} {} per {}".format(model_name, loss_name,
                                                                                           str(event.frequency)),
                                                            'name': str(event.phase),
                                                            'legend': [str(event.phase)]}})
                for loss_name, loss_value in monitors.items()]


class PlotMetrics(BaseVisdomHandler):
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END, Event.ON_TEST_BATCH_END,
                        Event.ON_BATCH_END]

    def __init__(self, visdom_logger: VisdomLogger, every=1):
        super().__init__(self.SUPPORTED_EVENTS, visdom_logger, every)

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        data = list()

        if self.should_handle(event):
            for model_name, monitor in monitors.items():
                data.extend(self.create_visdom_data(event, model_name, monitor[event.phase][Monitor.METRICS]))

        if data is not None:
            self.visdom_logger(data)

    def create_visdom_data(self, event, model_name, monitors):
        return [VisdomData(model_name, metric_name, PlotType.LINE_PLOT, event.frequency, [[event.iteration]],
                           [[metric_value]], params={'opts': {'xlabel': str(event.frequency), 'ylabel': metric_name,
                                                              'title': "{} {} per {}".format(model_name, metric_name,
                                                                                             str(event.frequency)),
                                                              'name': str(event.phase), 'legend': [str(event.phase)]}})
                for metric_name, metric_value in monitors.items()]


class PlotCustomVariables(BaseVisdomHandler):
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END, Event.ON_TRAIN_EPOCH_END, Event.ON_VALID_EPOCH_END, Event.ON_TEST_EPOCH_END,
                        Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END, Event.ON_TEST_BATCH_END, Event.ON_BATCH_END]

    def __init__(self, visdom_logger: VisdomLogger, variable_name, plot_type: PlotType, params, every=1):
        super().__init__(self.SUPPORTED_EVENTS, visdom_logger, every)
        self._variable_name = variable_name
        self._plot_type = plot_type
        self._params = params

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        data = None

        if self.should_handle(event):
            data = self.create_visdom_data(event, trainer)

        if data is not None:
            self.visdom_logger(data)

    def create_visdom_data(self, event: TemporalEvent, trainer):
        if self._plot_type == PlotType.LINE_PLOT and "name" not in self._params['opts'].keys():
            self._params['opts']['name'] = str(event.phase)

        return [VisdomData(trainer.name, self._variable_name, self._plot_type, event.frequency,
                           [event.iteration], trainer.custom_variables[self._variable_name], self._params)]


class PlotLR(BaseVisdomHandler):
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END, Event.ON_TEST_BATCH_END,
                        Event.ON_BATCH_END]

    def __init__(self, visdom_logger: VisdomLogger, every=1):
        super().__init__(self.SUPPORTED_EVENTS, visdom_logger, every)

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        data = None

        if self.should_handle(event):
            data = list(map(
                lambda model_trainer: self.create_visdom_data(event, model_trainer), trainer.model_trainers))

        if data is not None:
            self.visdom_logger(data)

    def create_visdom_data(self, event, model_trainer: ModelTrainer):
        return VisdomData(model_trainer.name, "Learning Rate", PlotType.LINE_PLOT, event.frequency, [event.iteration],
                          model_trainer.optimizer_lr,
                          params={'opts': {'xlabel': str(event.frequency),
                                           'ylabel': "Learning Rate",
                                           'title': "{} {} per {}".format(model_trainer.name, "Learning Rate",
                                                                          str(event.frequency)),
                                           'name': model_trainer.name,
                                           'legend': [model_trainer.name]}})


class PlotAvgGradientPerLayer(BaseVisdomHandler):
    SUPPORTED_EVENTS = [Event.ON_EPOCH_END, Event.ON_TRAIN_BATCH_END, Event.ON_VALID_BATCH_END, Event.ON_TEST_BATCH_END,
                        Event.ON_BATCH_END]

    def __init__(self, visdom_logger: VisdomLogger, every=1):
        super().__init__(self.SUPPORTED_EVENTS, visdom_logger, every)

    def __call__(self, event: TemporalEvent, monitors: dict, trainer: Trainer):
        data = None

        if self.should_handle(event):
            data = list(
                map(lambda model_trainer: self.create_visdom_data(event, model_trainer), trainer.model_trainers))

        if data is not None:
            self.visdom_logger(data)

    def create_visdom_data(self, event: TemporalEvent, model_trainer: ModelTrainer):
        avg_grads, layers = [], []

        for n, p in model_trainer.named_parameters():
            if p.requires_grad and ("bias" not in n):
                layers.append(n)
                if p.grad is not None:
                    avg_grads.append(p.grad.abs().mean().item())
                else:
                    avg_grads.append(0)

        return VisdomData(model_trainer.name, "Gradient Flow", PlotType.BAR_PLOT, event.frequency, y=layers,
                          x=avg_grads, params={'opts': {'xlabel': "Layers", 'ylabel': "Avg. Gradients",
                                                        'title': "{} {} per {}".format(model_trainer.name,
                                                                                       "Avg. Gradient", "Layer"),
                                                        'marginbottom': 200}})



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from typing import Union, Tuple

import torch
from enum import Enum
from ignite.metrics import Accuracy, Average, Precision, MeanAbsoluteError, MeanPairwiseDistance, MeanSquaredError, \
    Recall, RootMeanSquaredError, TopKCategoricalAccuracy, IoU, mIoU, Metric, ConfusionMatrix, MetricsLambda, \
    Fbeta, GeometricAverage

from kerosene.utils.constants import EPSILON
from kerosene.utils.tensors import flatten, to_onehot


class MetricType(Enum):
    Accuracy = "Accuracy"
    Average = "Average"
    ConfusionMatrix = "ConfusionMatrix"
    DiceCoefficient = "Dice"
    Fbeta = "Fbeta"
    GeneralizedDice = "GeneralizedDice"
    GeometricAverage = "GeometricAverage"
    IoU = "IoU"
    mIoU = "mIoU"
    MeanAbsoluteError = "MeanAbsoluteError"
    MeanPairwiseDistance = "MeanPairwiseDistance"
    MeanSquaredError = "MeanSquaredError"
    Precision = "Precision"
    Recall = "Recall"
    RootMeanSquaredError = "RootMeanSquaredError"
    TopKCategoricalAccuracy = "TopKCategoricalAccuracy"
    VariableAccumulation = "VariableAccumulation"

    def __str__(self):
        return self.value


class MetricFactory(object):
    def __init__(self):
        self._metrics = {
            "Accuracy": Accuracy,
            "Average": Average,
            "ConfusionMatrix": ConfusionMatrix,
            "Dice": Dice,
            "Fbeta": Fbeta,
            "GeneralizedDice": GeneralizedDice,
            "GeometricAverage": GeometricAverage,
            "IoU": IoU,
            "mIoU": mIoU,
            "MeanAbsoluteError": MeanAbsoluteError,
            "MeanPairwiseDistance": MeanPairwiseDistance,
            "MeanSquaredError": MeanSquaredError,
            "Precision": Precision,
            "Recall": Recall,
            "RootMeanSquaredError": RootMeanSquaredError,
            "TopKCategoricalAccuracy": TopKCategoricalAccuracy
        }

    def create(self, metric_type: Union[str, MetricType], params):
        return self._metrics[str(metric_type)](**params) if params is not None else self._metrics[str(metric_type)]()

    def register(self, metric: str, creator: Metric):
        """
        Add a new metric.
        Args:
           metric (str): Metric's name.
           creator: A torch or ignite module object wrapping the new custom metric function.
        """
        self._metrics[metric] = creator


class Dice(Metric):
    """
    The Dice Metric.
    """
    SUPPORTED_REDUCTIONS = [None, "mean"]

    def __init__(self, num_classes: int, reduction: Union[None, str] = "mean", average: str = None,
                 weight: torch.Tensor = None, ignore_index: int = -100,
                 output_transform: callable = lambda x: x) -> None:
        """
        Metric initializer.
        Args:
            num_classes (int): The number of classes in the problem. In case of images, num_classes should also count the background index 0.
            average (str, optional): Confusion matrix values averaging schema: None, "samples", "recall", "precision".
                Default is None. If `average="samples"` then confusion matrix values are normalized by the number of seen
                samples. If `average="recall"` then confusion matrix values are normalized such that diagonal values
                represent class recalls. If `average="precision"` then confusion matrix values are normalized such that
                diagonal values represent class precisions.
            reduction (str): The type of reduction to apply (e.g. 'mean').
            ignore_index (int, optional): To ignore an index in Dice computation.
            output_transform (callable, optional): a callable that is used to transform the
                output into the form expected by the metric. This can be useful if, for example, you have a multi-output model and
                you want to compute the metric with respect to one of the outputs.
        """
        if reduction not in self.SUPPORTED_REDUCTIONS:
            raise NotImplementedError("Reduction type not supported.")
        self._num_classes = num_classes
        self._ignore_index = ignore_index
        self._reduction = reduction
        self._weight = weight
        self._cm = ConfusionMatrix(num_classes=num_classes, average=average, output_transform=output_transform)
        self._metric = self.create_dice_metric(self._cm)
        super(Dice, self).__init__(output_transform=output_transform)

    def reset(self) -> None:
        """
        Reset the confusion matrix object.
        """
        self._cm.confusion_matrix = torch.zeros(self._num_classes, self._num_classes, dtype=torch.float64)

    def compute(self) -> torch.Tensor:
        """
        Compute the metric.
        Returns:
            :obj:`torch.Tensor`: The dice coefficient for each class.
        """
        return self._metric.compute()

    def update(self, output: Tuple[torch.Tensor, torch.Tensor]) -> None:
        """
        Update the confusion matrix with output values.
        Args:
            output (tuple of :obj:`torch.Tensor`): A tuple containing predictions and ground truth of the form `(y_pred, y)`.
        """
        self._cm.update(output)

    def create_dice_metric(self, cm: ConfusionMatrix):
        """
        Computes the Sørensen–Dice Coefficient (https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient)
        Args:
            cm (:obj:`ignite.metrics.ConfusionMatrix`): A confusion matrix representing the classification of data.
        Returns:
            array or float: The Sørensen–Dice Coefficient for each class or the mean Sørensen–Dice Coefficient.
        """
        # Increase floating point precision
        cm = cm.type(torch.float64)
        dice = 2 * cm.diag() / (cm.sum(dim=1) + cm.sum(dim=0) + EPSILON)

        if self._ignore_index != -100:
            def remove_index(dice_vector):
                try:
                    indices = list(range(len(dice_vector)))
                    indices.remove(self._ignore_index)
                    return dice_vector[indices]
                except ValueError as e:
                    raise IndexError(
                        "'ignore_index' must be non-negative, and lower than the number of classes in confusion matrix, but {} was given. ".format(
                            self._ignore_index))

            dice = MetricsLambda(remove_index, dice)

        if self._weight is not None:
            def multiply_weights(dice_vector):
                return self._weight * dice_vector

            dice = MetricsLambda(multiply_weights, dice)

        if self._reduction == "mean":
            dice = dice.mean()

        return dice


class GeneralizedDice(Metric):
    """
    The Generalized Dice Metric.
    """
    SUPPORTED_REDUCTIONS = [None, "mean"]

    def __init__(self, num_classes: int, reduction: str = None, average: str = None,
                 ignore_index: int = -100, output_transform: callable = lambda x: x) -> None:
        """
        Metric initializer.
        Args:
            num_classes (int): The number of classes in the problem. In case of images, num_classes should also count the background index 0.
            average (str, optional): Confusion matrix values averaging schema: None, "samples", "recall", "precision".
                Default is None. If `average="samples"` then confusion matrix values are normalized by the number of seen
                samples. If `average="recall"` then confusion matrix values are normalized such that diagonal values
                represent class recalls. If `average="precision"` then confusion matrix values are normalized such that
                diagonal values represent class precisions.
            reduction (str): The type of reduction to apply (e.g. 'mean').
            ignore_index (int, optional): To ignore an index in Dice computation.
            output_transform (callable, optional): a callable that is used to transform the
                output into the form expected by the metric. This can be useful if, for example, you have a multi-output model and
                you want to compute the metric with respect to one of the outputs.
        """
        if reduction not in self.SUPPORTED_REDUCTIONS:
            raise NotImplementedError("Reduction type not supported.")
        self._num_classes = num_classes
        self._average = average
        self._ignore_index = ignore_index
        self._metric = None
        self._reduction = reduction
        self._cm = ConfusionMatrix(num_classes=num_classes, average=average,
                                   output_transform=output_transform)
        super(GeneralizedDice, self).__init__(output_transform=output_transform)

    def reset(self) -> None:
        """
        Reset the confusion matrix object.
        """
        self._cm.confusion_matrix = torch.zeros(self._num_classes, self._num_classes, dtype=torch.float64)

    def compute(self) -> torch.Tensor:
        """
        Compute the metric.
        Returns:
            :obj:`torch.Tensor`: The dice coefficient for each class.
        """
        return self._metric.compute()

    def update(self, output: Tuple[torch.Tensor, torch.Tensor]) -> None:
        """
        Update the confusion matrix with output values.
        Args:
            output (tuple of :obj:`torch.Tensor`): A tuple containing predictions and ground truth of the form `(y_pred, y)`.
        """
        flattened_targets = flatten(to_onehot(output[1], output[0].size(1))).double()
        ones = torch.Tensor().new_ones((output[0].size(1),), dtype=torch.double, device=flattened_targets.device)
        weights = ones / torch.pow(flattened_targets.sum(-1), 2).clamp(min=EPSILON)

        self._metric = self.create_generalized_dice_metric(self._cm, weights)

        if self._reduction == "mean":
            self._metric = self._metric.mean()
        elif self._reduction is None:
            pass
        else:
            raise NotImplementedError("Reduction method not implemented.")

        self._cm.update(output)

    def create_generalized_dice_metric(self, cm: ConfusionMatrix, weight: torch.Tensor):
        """
        Computes the Sørensen–Dice Coefficient (https://en.wikipedia.org/wiki/Sørensen–Dice_coefficient)
        Args:
            cm (:obj:`ignite.metrics.ConfusionMatrix`): A confusion matrix representing the classification of data.
            weight (:obj:`torch.Tensor`): A weight vector which length equals to the number of classes.
        Returns:
            ignite.Metric: The Generalized Dice Coefficient Metric object.
        """

        # Increase floating point precision
        cm = cm.type(torch.float64)
        dice = 2 * (cm.diag() * weight) / (((cm.sum(dim=1) + cm.sum(dim=0)) * weight) + EPSILON)

        if self._ignore_index != -100:
            def remove_index(dice_vector):
                try:
                    indices = list(range(len(dice_vector)))
                    indices.remove(self._ignore_index)
                    return dice_vector[indices]
                except ValueError as e:
                    raise IndexError(
                        "'ignore_index' must be non-negative, and lower than the number of classes in confusion matrix, but {} was given. ".format(
                            self._ignore_index))

            return MetricsLambda(remove_index, dice)
        else:
            return dice


def to_tensor(value):
    if not isinstance(value, torch.Tensor):
        return torch.tensor([value])
    else:
        return value



from enum import Enum


class BaseStatus(Enum):

    def __str__(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, str):
            return self.value == other
        elif isinstance(other, BaseStatus):
            return self.value == other.value

    def __hash__(self):
        return hash(self.value)


class Status(BaseStatus):
    INITIALIZING = "Initializing"
    INITIALIZED = "Initialized"
    READY = "Ready"
    TRAINING = "Training"
    VALIDATING = "Validating"
    TESTING = "Testing"
    FINALIZED = "Finalized"



# -*- coding: utf-8 -*-
# Copyright 2019 Kerosene Authors. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import torch


def to_onehot(indices: torch.Tensor, num_classes: int) -> torch.Tensor:
    """
    Convert a tensor of indices of any shape `(N, ...)` to a tensor of one-hot indicators of shape
    `(N, num_classes, ...)`.

    Args:
        indices (:obj:`torch.Tensor`): Tensor of indices.
        num_classes (int): The number of classes of the problem.

    Returns:
        :obj:`torch.Tensor`: The one-hot representation of the input tensor.
    """
    onehot = torch.zeros(indices.shape[0], num_classes, *indices.shape[1:], device=indices.device)
    return onehot.scatter_(1, indices.unsqueeze(1), 1)


def flatten(tensor: torch.Tensor) -> torch.Tensor:
    """
    Flattens a given tensor such that the channel axis is first.
    The shapes are transformed as follows: (N, C, D, H, W) -> (C, N * D * H * W)

    Args:
        tensor (:obj:`torch.Tensor`): Tensor to flatten.

    Returns:
        :obj:`torch.Tensor`: The flattened, 1-D tensor.
    """
    C = tensor.size(1)
    # new axis order
    axis_order = (1, 0) + tuple(range(2, tensor.dim()))
    # Transpose: (N, C, D, H, W) -> (C, N, D, H, W)
    transposed = tensor.permute(axis_order).contiguous()
    # Flatten: (C, N, D, H, W) -> (C, N * D * H * W)
    return transposed.view(C, -1)



# -*- coding: utf-8 -*-
# Copyright 2019 Pierre-Luc Delisle. All Rights Reserved.
#
# Licensed under the MIT License;
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from typing import List

import torch
from torch.utils.data import DataLoader

from kerosene.configs.configs import RunConfiguration
from kerosene.training.trainers import ModelTrainer
from kerosene.training.trainers import Trainer
from kerosene.utils.devices import on_single_device


class MNISTTrainer(Trainer):

    def __init__(self, training_config, model_trainers: List[ModelTrainer],
                 train_data_loader: DataLoader, valid_data_loader: DataLoader, run_config: RunConfiguration):
        super(MNISTTrainer, self).__init__("MNISTTrainer", train_data_loader, valid_data_loader, model_trainers,
                                           run_config)

        self._training_config = training_config

    def train_step(self, inputs, target):
        model = self._model_trainers[0]

        pred = model.forward(inputs)
        model.compute_train_metric(pred, target)
        loss = model.compute_and_update_train_loss(pred, target)

        model.zero_grad()
        loss.backward()

        if not on_single_device(self._run_config.devices):
            self.average_gradients(model)

        model.step()

    def validate_step(self, inputs, target):
        model = self._model_trainers[0]

        pred = model.forward(inputs)
        model.compute_valid_metric(pred, target)
        model.compute_and_update_valid_loss(pred, target)

    def scheduler_step(self):
        self._model_trainers[0].scheduler_step()

    @staticmethod
    def average_gradients(model):
        size = float(torch.distributed.get_world_size())
        for param in model.parameters():
            torch.distributed.all_reduce(param.grad.data, op=torch.distributed.ReduceOp.SUM)
            param.grad.data /= size

    def on_epoch_begin(self):
        pass

    def on_epoch_end(self):
        pass
